#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

import asyncio
import logging

import uamqp
from uamqp import c_uamqp, connection
from uamqp.async_ops.utils import get_dict_with_loop_if_needed

_logger = logging.getLogger(__name__)


class ConnectionAsync(connection.Connection):
    """An Asynchronous AMQP Connection. A single Connection can have multiple
    Sessions, and can be shared between multiple Clients.

    :ivar max_frame_size: Maximum AMQP frame size. Default is 63488 bytes.
    :vartype max_frame_size: int
    :ivar channel_max: Maximum number of Session channels in the Connection.
    :vartype channel_max: int
    :ivar idle_timeout: Timeout in milliseconds after which the Connection will close
     if there is no further activity.
    :vartype idle_timeout: int
    :ivar properties: Connection properties.
    :vartype properties: dict

    :param hostname: The hostname of the AMQP service with which to establish
     a connection.
    :type hostname: bytes or str
    :param sasl: Authentication for the connection. If none is provided SASL Annoymous
     authentication will be used.
    :type sasl: ~uamqp.authentication.common.AMQPAuth
    :param container_id: The name for the client, also known as the Container ID.
     If no name is provided, a random GUID will be used.
    :type container_id: str or bytes
    :param max_frame_size: Maximum AMQP frame size. Default is 63488 bytes.
    :type max_frame_size: int
    :param channel_max: Maximum number of Session channels in the Connection.
    :type channel_max: int
    :param idle_timeout: Timeout in milliseconds after which the Connection will close
     if there is no further activity.
    :type idle_timeout: int
    :param properties: Connection properties.
    :type properties: dict
    :param remote_idle_timeout_empty_frame_send_ratio: Ratio of empty frames to
     idle time for Connections with no activity. Value must be between
     0.0 and 1.0 inclusive. Default is 0.5.
    :type remote_idle_timeout_empty_frame_send_ratio: float
    :param debug: Whether to turn on network trace logs. If `True`, trace logs
     will be logged at INFO level. Default is `False`.
    :type debug: bool
    :param encoding: The encoding to use for parameters supplied as strings.
     Default is 'UTF-8'
    :type encoding: str
    """

    def __init__(self, hostname, sasl,
                 container_id=False,
                 max_frame_size=None,
                 channel_max=None,
                 idle_timeout=None,
                 properties=None,
                 remote_idle_timeout_empty_frame_send_ratio=None,
                 error_policy=None,
                 debug=False,
                 encoding='UTF-8',
                 loop=None):
        self._internal_kwargs = get_dict_with_loop_if_needed(loop)
        super(ConnectionAsync, self).__init__(
            hostname, sasl,
            container_id=container_id,
            max_frame_size=max_frame_size,
            channel_max=channel_max,
            idle_timeout=idle_timeout,
            properties=properties,
            remote_idle_timeout_empty_frame_send_ratio=remote_idle_timeout_empty_frame_send_ratio,
            error_policy=error_policy,
            debug=debug,
            encoding=encoding)
        self._async_lock = asyncio.Lock(**self._internal_kwargs)

    async def __aenter__(self):
        """Open the Connection in an async context manager."""
        return self

    async def __aexit__(self, *args):
        """Close the Connection when exiting an async context manager."""
        _logger.debug("Exiting connection %r context.", self.container_id)
        await self.destroy_async()
        _logger.debug("Finished exiting connection %r context.", self.container_id)

    async def _close_async(self):
        _logger.info("Shutting down connection %r.", self.container_id)
        self._closing = True
        if self._cbs:
            await self.auth.close_authenticator_async()
            self._cbs = None
        self._conn.destroy()
        self.auth.close()
        _logger.info("Connection shutdown complete %r.", self.container_id)

    @property
    def loop(self):
        return self._internal_kwargs.get("loop")

    async def lock_async(self, timeout=3.0):
        await asyncio.wait_for(self._async_lock.acquire(), timeout=timeout, **self._internal_kwargs)

    def release_async(self):
        try:
            self._async_lock.release()
        except RuntimeError:
            pass
        except:
            _logger.debug("Got error when attempting to release async connection lock.")
            try:
                self._async_lock.release()
            except RuntimeError:
                pass
            raise

    async def work_async(self):
        """Perform a single Connection iteration asynchronously."""
        try:
            raise self._error
        except TypeError:
            pass
        except Exception as e:
            _logger.warning("%r", e)
            raise
        try:
            await self.lock_async()
            if self._closing:
                _logger.debug("Connection unlocked but shutting down.")
                return
            await asyncio.sleep(0, **self._internal_kwargs)
            self._conn.do_work()
        except asyncio.TimeoutError:
            _logger.debug("Connection %r timed out while waiting for lock acquisition.", self.container_id)
        finally:
            await asyncio.sleep(0, **self._internal_kwargs)
            self.release_async()

    async def sleep_async(self, seconds):
        """Lock the connection for a given number of seconds.

        :param seconds: Length of time to lock the connection.
        :type seconds: int
        """
        try:
            await self.lock_async()
            await asyncio.sleep(seconds, **self._internal_kwargs)
        except asyncio.TimeoutError:
            _logger.debug("Connection %r timed out while waiting for lock acquisition.", self.container_id)
        finally:
            self.release_async()

    async def redirect_async(self, redirect_error, auth):
        """Redirect the connection to an alternative endpoint.
        :param redirect: The Link DETACH redirect details.
        :type redirect: ~uamqp.errors.LinkRedirect
        :param auth: Authentication credentials to the redirected endpoint.
        :type auth: ~uamqp.authentication.common.AMQPAuth
        """
        _logger.info("Redirecting connection %r.", self.container_id)
        try:
            await self.lock_async()
            if self.hostname == redirect_error.hostname:
                return
            if self._state != c_uamqp.ConnectionState.END:
                await self._close_async()
            self.hostname = redirect_error.hostname
            self.auth = auth
            self._conn = self._create_connection(auth)
            for setting, value in self._settings.items():
                setattr(self, setting, value)
            self._error = None
            self._closing = False
        except asyncio.TimeoutError:
            _logger.debug("Connection %r timed out while waiting for lock acquisition.", self.container_id)
        finally:
            self.release_async()

    async def destroy_async(self):
        """Close the connection asynchronously, and close any associated
        CBS authentication session.
        """
        try:
            await self.lock_async()
            _logger.debug("Unlocked connection %r to close.", self.container_id)
            await self._close_async()
        except asyncio.TimeoutError:
            _logger.debug(
                "Connection %r timed out while waiting for lock acquisition on destroy. Destroying anyway.",
                self.container_id)
            await self._close_async()
        finally:
            self.release_async()
        uamqp._Platform.deinitialize()  # pylint: disable=protected-access



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

# Python imports
from enum import Enum
import logging
import uuid
import copy

import six

# C improts
from libc cimport stdint
from libc.stdlib cimport malloc, realloc, free
from libc.string cimport memcpy

cimport cython
cimport c_amqpvalue
cimport c_amqp_definitions


_logger = logging.getLogger(__name__)


cdef int encode_bytes_callback(void* context, const unsigned char* encoded_bytes, size_t length):
    context_obj = <object>context
    context_obj.append(encoded_bytes[:length])
    return 0


cdef get_amqp_value_type(c_amqpvalue.AMQP_VALUE value):
    type_val = c_amqpvalue.amqpvalue_get_type(value)
    try:
        return AMQPType(type_val)
    except ValueError:
        _logger.info("Received unrecognized type value: %r", type_val)
        return AMQPType.ErrorType


cpdef enocde_batch_value(AMQPValue value, message_body):
    if c_amqpvalue.amqpvalue_encode(<c_amqpvalue.AMQP_VALUE>value._c_value,
                                     <c_amqpvalue.AMQPVALUE_ENCODER_OUTPUT>encode_bytes_callback,
                                     <void*>message_body) != 0:
        raise ValueError("Failed to encode batched message data.")


class AMQPType(Enum):
    NullValue = c_amqpvalue.AMQP_TYPE_TAG.AMQP_TYPE_NULL
    BoolValue = c_amqpvalue.AMQP_TYPE_TAG.AMQP_TYPE_BOOL
    UByteValue = c_amqpvalue.AMQP_TYPE_TAG.AMQP_TYPE_UBYTE
    UShortValue = c_amqpvalue.AMQP_TYPE_TAG.AMQP_TYPE_USHORT
    UIntValue = c_amqpvalue.AMQP_TYPE_TAG.AMQP_TYPE_UINT
    ULongValue = c_amqpvalue.AMQP_TYPE_TAG.AMQP_TYPE_ULONG
    ByteValue = c_amqpvalue.AMQP_TYPE_TAG.AMQP_TYPE_BYTE
    ShortValue = c_amqpvalue.AMQP_TYPE_TAG.AMQP_TYPE_SHORT
    IntValue = c_amqpvalue.AMQP_TYPE_TAG.AMQP_TYPE_INT
    LongValue = c_amqpvalue.AMQP_TYPE_TAG.AMQP_TYPE_LONG
    FloatValue = c_amqpvalue.AMQP_TYPE_TAG.AMQP_TYPE_FLOAT
    DoubleValue = c_amqpvalue.AMQP_TYPE_TAG.AMQP_TYPE_DOUBLE
    CharValue = c_amqpvalue.AMQP_TYPE_TAG.AMQP_TYPE_CHAR
    TimestampValue = c_amqpvalue.AMQP_TYPE_TAG.AMQP_TYPE_TIMESTAMP
    UUIDValue = c_amqpvalue.AMQP_TYPE_TAG.AMQP_TYPE_UUID
    BinaryValue = c_amqpvalue.AMQP_TYPE_TAG.AMQP_TYPE_BINARY
    StringValue = c_amqpvalue.AMQP_TYPE_TAG.AMQP_TYPE_STRING
    SymbolValue = c_amqpvalue.AMQP_TYPE_TAG.AMQP_TYPE_SYMBOL
    ListValue = c_amqpvalue.AMQP_TYPE_TAG.AMQP_TYPE_LIST
    DictValue = c_amqpvalue.AMQP_TYPE_TAG.AMQP_TYPE_MAP
    ArrayValue = c_amqpvalue.AMQP_TYPE_TAG.AMQP_TYPE_ARRAY
    DescribedType = c_amqpvalue.AMQP_TYPE_TAG.AMQP_TYPE_DESCRIBED
    CompositeType = c_amqpvalue.AMQP_TYPE_TAG.AMQP_TYPE_COMPOSITE
    UnknownType = c_amqpvalue.AMQP_TYPE_TAG.AMQP_TYPE_UNKNOWN
    ErrorType = 999


cdef value_factory(c_amqpvalue.AMQP_VALUE value):
    type_val = get_amqp_value_type(value)
    _logger.debug("Wrapping value type: %r", type_val)
    if type_val == AMQPType.NullValue:
        new_obj = AMQPValue()
    elif type_val == AMQPType.BoolValue:
        new_obj = BoolValue()
    elif type_val == AMQPType.UByteValue:
        new_obj = UByteValue()
    elif type_val == AMQPType.UShortValue:
        new_obj = UShortValue()
    elif type_val == AMQPType.UIntValue:
        new_obj = UIntValue()
    elif type_val == AMQPType.ULongValue:
        new_obj = ULongValue()
    elif type_val == AMQPType.ByteValue:
        new_obj = ByteValue()
    elif type_val == AMQPType.ShortValue:
        new_obj = ShortValue()
    elif type_val == AMQPType.IntValue:
        new_obj = IntValue()
    elif type_val == AMQPType.LongValue:
        new_obj = LongValue()
    elif type_val == AMQPType.FloatValue:
        new_obj = FloatValue()
    elif type_val == AMQPType.DoubleValue:
        new_obj = DoubleValue()
    elif type_val == AMQPType.CharValue:
        new_obj = CharValue()
    elif type_val == AMQPType.TimestampValue:
        new_obj = TimestampValue()
    elif type_val == AMQPType.UUIDValue:
        new_obj = UUIDValue()
    elif type_val == AMQPType.BinaryValue:
        new_obj = BinaryValue()
    elif type_val == AMQPType.StringValue:
        new_obj = StringValue()
    elif type_val == AMQPType.SymbolValue:
        new_obj = SymbolValue()
    elif type_val == AMQPType.ListValue:
        new_obj = ListValue()
    elif type_val == AMQPType.DictValue:
        new_obj = DictValue()
    elif type_val == AMQPType.ArrayValue:  # Do checking
        new_obj = ArrayValue()
    elif type_val == AMQPType.CompositeType:
        new_obj = CompositeValue()
    elif type_val == AMQPType.DescribedType:
        new_obj = DescribedValue()
    else:
        error = "Unrecognized AMQPType: {}".format(type_val)
        _logger.info(error)
        raise TypeError(error)
    new_obj.wrap(value)
    return new_obj


cpdef null_value():
    new_obj = AMQPValue()
    new_obj.create()
    return new_obj


cpdef bool_value(bint value):
    value = 1 if value else 0
    new_obj = BoolValue()
    new_obj.create(value)
    return new_obj


cpdef ubyte_value(unsigned char value):
    new_obj = UByteValue()
    new_obj.create(value)
    return new_obj


cpdef ushort_value(stdint.uint16_t value):
    new_obj = UShortValue()
    new_obj.create(value)
    return new_obj


cpdef uint_value(stdint.uint32_t value):
    new_obj = UIntValue()
    new_obj.create(value)
    return new_obj


cpdef ulong_value(stdint.uint64_t value):
    new_obj = ULongValue()
    new_obj.create(value)
    return new_obj


cpdef byte_value(char value):
    new_obj = ByteValue()
    new_obj.create(value)
    return new_obj


cpdef short_value(stdint.int16_t value):
    new_obj = ShortValue()
    new_obj.create(value)
    return new_obj


cpdef int_value(stdint.int32_t value):
    new_obj = IntValue()
    new_obj.create(value)
    return new_obj


cpdef long_value(stdint.int64_t value):
    new_obj = LongValue()
    new_obj.create(value)
    return new_obj


cpdef float_value(float value):
    new_obj = FloatValue()
    new_obj.create(value)
    return new_obj


cpdef double_value(double value):
    new_obj = DoubleValue()
    new_obj.create(value)
    return new_obj


cpdef char_value(stdint.uint32_t value):
    new_obj = CharValue()
    new_obj.create(value)
    return new_obj


cpdef timestamp_value(stdint.int64_t value):
    new_obj = TimestampValue()
    new_obj.create(value)
    return new_obj


cpdef uuid_value(value):
    if not isinstance(value, uuid.UUID):
        raise TypeError("Input value must be type UUID.")
    new_obj = UUIDValue()
    new_obj.create(<bytes>value.bytes)
    return new_obj


cpdef binary_value(value):
    bytes_value = six.binary_type(value)
    new_obj = BinaryValue()
    new_obj.create(<bytes>bytes_value)
    return new_obj


cpdef string_value(char* value):
    new_obj = StringValue()
    new_obj.create(value)
    return new_obj


cpdef symbol_value(char* value):
    new_obj = SymbolValue()
    new_obj.create(value)
    return new_obj


cpdef list_value():
    new_obj = ListValue()
    new_obj.create()
    return new_obj


cpdef dict_value():
    new_obj = DictValue()
    new_obj.create()
    return new_obj


cpdef array_value():
    new_obj = ArrayValue()
    new_obj.create()
    return new_obj


cpdef described_value(AMQPValue descriptor, AMQPValue value):
    new_obj = DescribedValue()
    new_obj.create(descriptor, value)
    return new_obj


cdef class AMQPValue(StructBase):

    _type = AMQPType.NullValue
    cdef c_amqpvalue.AMQP_VALUE _c_value

    def __cinit__(self):
        pass

    def __dealloc__(self):
        if _logger:
            _logger.debug("Deallocating %r", self.__class__.__name__)
        self.destroy()

    def __eq__(self, AMQPValue other):
        return c_amqpvalue.amqpvalue_are_equal(self._c_value, other._c_value)

    def __ne__(self, AMQPValue other):
        return not c_amqpvalue.amqpvalue_are_equal(self._c_value, other._c_value)

    def __bytes__(self):
        return self._as_string()

    def __str__(self):
        as_bytes = self._as_string()
        if six.PY3:
            try:
                return as_bytes.decode('UTF-8')
            except UnicodeDecodeError:
                pass
        return str(as_bytes)

    def __unicode__(self):
        as_bytes = self._as_string()
        try:
            return six.text_type(as_bytes.decode('UTF-8'))
        except UnicodeDecodeError:
            return six.text_type(as_bytes)

    cpdef _as_string(self):
        cdef c_amqpvalue.AMQP_VALUE value
        cdef const char* as_string
        value = c_amqpvalue.amqpvalue_clone(self._c_value)
        if <void*>value == NULL:
            self._value_error()
        as_string = c_amqpvalue.amqpvalue_to_string(value)
        py_string = <bytes> as_string
        free(as_string)
        c_amqpvalue.amqpvalue_destroy(self._c_value)
        return py_string

    cdef _validate(self):
        if <void*>self._c_value is NULL:
            self._memory_error()

    cpdef destroy(self):
        try:
            if <void*>self._c_value is not NULL:
                if _logger:
                    _logger.debug("Destroying %r", self.__class__.__name__)
                c_amqpvalue.amqpvalue_destroy(self._c_value)
        except KeyboardInterrupt:
            pass
        finally:
            self._c_value = <c_amqpvalue.AMQP_VALUE>NULL

    cdef wrap(self, c_amqpvalue.AMQP_VALUE value):
        self.destroy()
        self._c_value = value
        self._validate()

    def create(self):
        new_value = c_amqpvalue.amqpvalue_create_null()
        self.wrap(new_value)

    @property
    def type(self):
        if <void*>self._c_value is NULL:
            self._null_error()
        obj_type = get_amqp_value_type(self._c_value)
        return obj_type

    @property
    def value(self):
        return None

    cpdef get_encoded_size(self):
        cdef size_t length
        if c_amqpvalue.amqpvalue_get_encoded_size(self._c_value, &length) != 0:
            self._value_error("Failed to get encoded size.")
        return length

    cpdef clone(self):
        cdef c_amqpvalue.AMQP_VALUE value
        value = c_amqpvalue.amqpvalue_clone(self._c_value)
        if <void*>value == NULL:
            self._value_error()
        return value_factory(value)


cdef class BoolValue(AMQPValue):

    _type = AMQPType.BoolValue

    cpdef _bool_value(self):
        cdef bint _value
        if c_amqpvalue.amqpvalue_get_boolean(self._c_value, &_value) == 0:
            return _value
        else:
            self._value_error()

    def create(self, bint value):
        new_value = c_amqpvalue.amqpvalue_create_boolean(value)
        self.wrap(new_value)

    @property
    def value(self):
        assert self.type
        str_value = str(self)
        if str_value in ["false"]:
            return False
        elif str_value in ["true"]:
            return True
        else:
            self._value_error()


cdef class UByteValue(AMQPValue):

    _type = AMQPType.UByteValue

    def create(self, unsigned char value):
        new_value = c_amqpvalue.amqpvalue_create_ubyte(value)
        self.wrap(new_value)

    @property
    def value(self):
        assert self.type
        cdef unsigned char _value
        if c_amqpvalue.amqpvalue_get_ubyte(self._c_value, &_value) == 0:
            return _value
        else:
            self._value_error()


cdef class UShortValue(AMQPValue):

    _type = AMQPType.UShortValue

    def create(self, stdint.uint16_t value):
        new_value = c_amqpvalue.amqpvalue_create_ushort(value)
        self.wrap(new_value)

    @property
    def value(self):
        assert self.type
        cdef stdint.uint16_t _value
        if c_amqpvalue.amqpvalue_get_ushort(self._c_value, &_value) == 0:
            return _value
        else:
            self._value_error()


cdef class UIntValue(AMQPValue):

    _type = AMQPType.UIntValue

    def create(self, stdint.uint32_t value):
        new_value = c_amqpvalue.amqpvalue_create_uint(value)
        self.wrap(new_value)

    @property
    def value(self):
        assert self.type
        cdef stdint.uint32_t _value
        if c_amqpvalue.amqpvalue_get_uint(self._c_value, &_value) == 0:
            return _value
        else:
            self._value_error()


cdef class ULongValue(AMQPValue):

    _type = AMQPType.ULongValue

    def create(self, stdint.uint64_t value):
        new_value = c_amqpvalue.amqpvalue_create_ulong(value)
        self.wrap(new_value)

    @property
    def value(self):
        assert self.type
        cdef stdint.uint64_t _value
        if c_amqpvalue.amqpvalue_get_ulong(self._c_value, &_value) == 0:
            return _value
        else:
            self._value_error()


cdef class ByteValue(AMQPValue):

    _type = AMQPType.ByteValue

    def create(self, char value):
        new_value = c_amqpvalue.amqpvalue_create_byte(value)
        self.wrap(new_value)

    @property
    def value(self):
        assert self.type
        cdef char _value
        if c_amqpvalue.amqpvalue_get_byte(self._c_value, &_value) == 0:
            return _value
        else:
            self._value_error()


cdef class ShortValue(AMQPValue):

    _type = AMQPType.ShortValue

    def create(self, stdint.int16_t value):
        new_value = c_amqpvalue.amqpvalue_create_short(value)
        self.wrap(new_value)

    @property
    def value(self):
        assert self.type
        cdef stdint.int16_t _value
        if c_amqpvalue.amqpvalue_get_short(self._c_value, &_value) == 0:
            return _value
        else:
            self._value_error()


cdef class IntValue(AMQPValue):

    _type = AMQPType.IntValue

    def create(self, stdint.int32_t value):
        new_value = c_amqpvalue.amqpvalue_create_int(value)
        self.wrap(new_value)

    @property
    def value(self):
        assert self.type
        cdef stdint.int32_t _value
        if c_amqpvalue.amqpvalue_get_int(self._c_value, &_value) == 0:
            return _value
        else:
            self._value_error()


cdef class LongValue(AMQPValue):

    _type = AMQPType.LongValue

    def create(self, stdint.int64_t value):
        new_value = c_amqpvalue.amqpvalue_create_long(value)
        self.wrap(new_value)

    @property
    def value(self):
        assert self.type
        cdef stdint.int64_t _value
        if c_amqpvalue.amqpvalue_get_long(self._c_value, &_value) == 0:
            return _value
        else:
            self._value_error()


cdef class FloatValue(AMQPValue):

    _type = AMQPType.FloatValue

    def create(self, float value):
        new_value = c_amqpvalue.amqpvalue_create_float(value)
        self.wrap(new_value)

    @property
    def value(self):
        assert self.type
        cdef float _value
        if c_amqpvalue.amqpvalue_get_float(self._c_value, &_value) == 0:
            return _value
        else:
            self._value_error()


cdef class DoubleValue(AMQPValue):

    _type = AMQPType.DoubleValue

    def create(self, double value):
        new_value = c_amqpvalue.amqpvalue_create_double(value)
        self.wrap(new_value)

    @property
    def value(self):
        assert self.type
        cdef double _value
        if c_amqpvalue.amqpvalue_get_double(self._c_value, &_value) == 0:
            return _value
        else:
            self._value_error()


cdef class CharValue(AMQPValue):

    _type = AMQPType.CharValue

    def create(self, stdint.uint32_t value):
        new_value = c_amqpvalue.amqpvalue_create_char(value)
        self.wrap(new_value)

    @property
    def value(self):
        assert self.type
        cdef stdint.uint32_t _value
        if c_amqpvalue.amqpvalue_get_char(self._c_value, &_value) == 0:
            return chr(_value)
        else:
            self._value_error()


cdef class TimestampValue(AMQPValue):

    _type = AMQPType.TimestampValue

    def create(self, stdint.int64_t value):
        new_value = c_amqpvalue.amqpvalue_create_timestamp(value)
        self.wrap(new_value)

    @property
    def value(self):
        assert self.type
        cdef stdint.int64_t _value
        if c_amqpvalue.amqpvalue_get_timestamp(self._c_value, &_value) == 0:
            return _value
        else:
            self._value_error()


cdef class UUIDValue(AMQPValue):

    _type = AMQPType.UUIDValue

    def create(self, bytes value):
        new_value = c_amqpvalue.amqpvalue_create_uuid(value)
        self.wrap(new_value)

    @property
    def value(self):
        assert self.type
        str_val = str(self)
        return uuid.UUID(str_val)  # TODO: Get proper value


cdef class BinaryValue(AMQPValue):

    _type = AMQPType.BinaryValue

    def create(self, bytes value):
        cdef c_amqpvalue.amqp_binary _binary
        length = len(list(value))
        _binary.length = length
        _binary.bytes = <char*>value
        new_value = c_amqpvalue.amqpvalue_create_binary(_binary)
        self.wrap(new_value)

    def __len__(self):
        assert self.type
        cdef c_amqpvalue.amqp_binary _value
        if c_amqpvalue.amqpvalue_get_binary(self._c_value, &_value) == 0:
            return _value.length
        else:
            self._value_error()

    @property
    def value(self):
        assert self.type
        cdef c_amqpvalue.amqp_binary _value
        cdef char* bytes_value
        if c_amqpvalue.amqpvalue_get_binary(self._c_value, &_value) == 0:
            bytes_value = <char*>_value.bytes
            return bytes_value[:_value.length]
        else:
            self._value_error()


cdef class StringValue(AMQPValue):

    _type = AMQPType.StringValue

    def create(self, char* value):
        new_value = c_amqpvalue.amqpvalue_create_string(value)
        self.wrap(new_value)

    @property
    def value(self):
        assert self.type
        cdef const char* _value
        if c_amqpvalue.amqpvalue_get_string(self._c_value, &_value) == 0:
            return copy.deepcopy(_value)
        else:
            self._value_error()


cdef class SymbolValue(AMQPValue):

    _type = AMQPType.SymbolValue

    def create(self, char* value):
        new_value = c_amqpvalue.amqpvalue_create_symbol(value)
        self.wrap(new_value)

    @property
    def value(self):
        assert self.type
        cdef const char* _value
        if c_amqpvalue.amqpvalue_get_symbol(self._c_value, &_value) == 0:
            return copy.deepcopy(_value)
        else:
            self._value_error()


cdef class ListValue(AMQPValue):

    _type = AMQPType.ListValue

    def create(self):
        new_value = c_amqpvalue.amqpvalue_create_list()
        self.wrap(new_value)

    def __len__(self):
        return self.size

    def __getitem__(self, size_t index):
        if index >= self.size:
            raise IndexError("Index is out of range.")
        cdef c_amqpvalue.AMQP_VALUE value
        value = c_amqpvalue.amqpvalue_get_list_item(self._c_value, index)
        if <void*>value == NULL:
            self._value_error()
        try:
            return value_factory(value)
        except TypeError:
            return None

    def __setitem__(self, stdint.uint32_t index, AMQPValue value):
        assert value.type
        if index >= self.size:
            raise IndexError("Index is out of range.")
        if c_amqpvalue.amqpvalue_set_list_item(self._c_value, index, value._c_value) != 0:
            self._value_error()

    @property
    def size(self):
        assert self.type
        cdef stdint.uint32_t _value
        if c_amqpvalue.amqpvalue_get_list_item_count(self._c_value, &_value) == 0:
            return _value
        else:
            self._value_error()

    @size.setter
    def size(self, stdint.uint32_t length):
        assert self.type
        if c_amqpvalue.amqpvalue_set_list_item_count(self._c_value, length) != 0:
            self._value_error()

    @property
    def value(self):
        assert self.type
        value = []
        for i in range(self.size):
            value.append(copy.deepcopy(self[i].value))
        return value


cdef class DictValue(AMQPValue):

    _type = AMQPType.DictValue

    def create(self):
        new_value = c_amqpvalue.amqpvalue_create_map()
        self.wrap(new_value)

    def __len__(self):
        return self.size

    def __getitem__(self, AMQPValue key):
        cdef c_amqpvalue.AMQP_VALUE value
        value = c_amqpvalue.amqpvalue_get_map_value(self._c_value, key._c_value)
        if <void*>value == NULL:
            raise KeyError("No value found for key: {}".format(key.value))
        try:
            return value_factory(value)
        except TypeError:
            return None

    def __setitem__(self, AMQPValue key, AMQPValue value):
        if c_amqpvalue.amqpvalue_set_map_value(self._c_value, key._c_value, value._c_value) != 0:
            self._value_error()

    def get(self, stdint.uint32_t index):
        if index >= self.size:
            raise IndexError("Index is out of range.")
        cdef c_amqpvalue.AMQP_VALUE key
        cdef c_amqpvalue.AMQP_VALUE value
        if c_amqpvalue.amqpvalue_get_map_key_value_pair(self._c_value, index, &key,  &value) != 0:
            self._value_error()
        return (value_factory(key), value_factory(value))

    @property
    def size(self):
        assert self.type
        cdef stdint.uint32_t _value
        if c_amqpvalue.amqpvalue_get_map_pair_count(self._c_value, &_value) == 0:
            return _value
        else:
            self._value_error()

    @property
    def value(self):
        assert self.type
        value = {}
        for i in range(self.size):
            key, item = self.get(i)
            value[copy.deepcopy(key.value)] = copy.deepcopy(item.value)
        return value


cdef class ArrayValue(AMQPValue):

    _type = AMQPType.ArrayValue

    def create(self):
        self._c_value = c_amqpvalue.amqpvalue_create_array()
        self._validate()

    def __len__(self):
        return self.size

    def __getitem__(self, stdint.uint32_t index):
        if index >= self.size:
            raise IndexError("Index is out of range.")
        cdef c_amqpvalue.AMQP_VALUE value
        value = c_amqpvalue.amqpvalue_get_array_item(self._c_value, index)
        if <void*>value == NULL:
            self._value_error()
        return value_factory(value)

    @property
    def size(self):
        assert self.type
        cdef stdint.uint32_t _value
        if c_amqpvalue.amqpvalue_get_array_item_count(self._c_value, &_value) == 0:
            return _value
        else:
            self._value_error()

    cpdef append(self, AMQPValue value):
        assert self.type
        cdef c_amqpvalue.AMQP_VALUE _value
        _value = <c_amqpvalue.AMQP_VALUE>value._c_value
        if c_amqpvalue.amqpvalue_add_array_item(self._c_value, <c_amqpvalue.AMQP_VALUE>value._c_value) != 0:
            self._value_error()

    @property
    def value(self):
        assert self.type
        value = []
        for i in range(self.size):
            value.append(copy.deepcopy(self[i].value))
        return value


cdef class CompositeValue(AMQPValue):

    _type = AMQPType.CompositeType

    def create(self, AMQPValue descriptor, stdint.uint32_t list_size):
        new_value = c_amqpvalue.amqpvalue_create_composite(descriptor._c_value, list_size)
        self.wrap(new_value)

    def create_from_long(self, stdint.uint64_t descriptor):
        new_value = c_amqpvalue.amqpvalue_create_composite_with_ulong_descriptor(descriptor)
        self.wrap(new_value)

    def __len__(self):
        return self.size

    def __getitem__(self, stdint.uint32_t index):
        if index >= self.size:
            raise IndexError("Index is out of range.")
        cdef c_amqpvalue.AMQP_VALUE value
        value = c_amqpvalue.amqpvalue_get_composite_item(self._c_value, index)
        if <void*>value == NULL:
            self._value_error()
        try:
            return value_factory(value)
        except TypeError:
            raise IndexError("No item found at index {}".format(index))

    def __setitem__(self, stdint.uint32_t index, AMQPValue value):
        assert value.type
        if index >= self.size:
            raise IndexError("Index is out of range.")
        if c_amqpvalue.amqpvalue_set_composite_item(self._c_value, index, value._c_value) != 0:
            self._value_error()

    @property
    def size(self):
        assert self.type
        cdef stdint.uint32_t _value
        if c_amqpvalue.amqpvalue_get_composite_item_count(self._c_value, &_value) == 0:
            return _value
        else:
            self._value_error()

    def pop(self, stdint.uint32_t index):
        if index >= self.size:
            raise IndexError("Index is out of range.")
        cdef c_amqpvalue.AMQP_VALUE value
        value = c_amqpvalue.amqpvalue_get_composite_item(self._c_value, index)
        if <void*>value == NULL:
            self._value_error()
        try:
            return value_factory(value)
        except TypeError:
            raise IndexError("No item found at index {}".format(index))


cdef class DescribedValue(AMQPValue):

    _type = AMQPType.DescribedType

    def create(self, AMQPValue descriptor, AMQPValue value):
        cdef c_amqpvalue.AMQP_VALUE cloned_descriptor
        cdef c_amqpvalue.AMQP_VALUE cloned_value
        cloned_descriptor = c_amqpvalue.amqpvalue_clone(<c_amqpvalue.AMQP_VALUE>descriptor._c_value)
        if <void*>cloned_descriptor == NULL:
            self._value_error()
        cloned_value = c_amqpvalue.amqpvalue_clone(<c_amqpvalue.AMQP_VALUE>value._c_value)
        if <void*>cloned_value == NULL:
            self._value_error()
        new_value = c_amqpvalue.amqpvalue_create_described(cloned_descriptor, cloned_value)
        self.wrap(new_value)

    @property
    def description(self):
        cdef c_amqpvalue.AMQP_VALUE value
        cdef c_amqpvalue.AMQP_VALUE cloned
        value = c_amqpvalue.amqpvalue_get_inplace_descriptor(self._c_value)
        if <void*>value == NULL:
            self._value_error()
        cloned = c_amqpvalue.amqpvalue_clone(value)
        if <void*>cloned == NULL:
            self._value_error()
        return value_factory(cloned)

    @property
    def data(self):
        cdef c_amqpvalue.AMQP_VALUE value
        cdef c_amqpvalue.AMQP_VALUE cloned
        value = c_amqpvalue.amqpvalue_get_inplace_described_value(self._c_value)
        if <void*>value == NULL:
            self._value_error()
        cloned = c_amqpvalue.amqpvalue_clone(value)
        if <void*>cloned == NULL:
            self._value_error()
        return value_factory(cloned)

    @property
    def value(self):
        assert self.type
        #descriptor = self.description
        described = self.data
        return copy.deepcopy(described.value)



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

from libc cimport stdint

cimport c_amqp_definitions
cimport c_amqpvalue
cimport c_xio

cdef extern from "azure_uamqp_c/connection.h":

    ctypedef struct CONNECTION_HANDLE:
        pass
    ctypedef struct ENDPOINT_HANDLE:
        pass
    ctypedef struct ON_CONNECTION_CLOSED_EVENT_SUBSCRIPTION_HANDLE:
        pass

    cdef enum CONNECTION_STATE_TAG:
        CONNECTION_STATE_START,
        CONNECTION_STATE_HDR_RCVD,
        CONNECTION_STATE_HDR_SENT,
        CONNECTION_STATE_HDR_EXCH,
        CONNECTION_STATE_OPEN_PIPE,
        CONNECTION_STATE_OC_PIPE,
        CONNECTION_STATE_OPEN_RCVD,
        CONNECTION_STATE_OPEN_SENT,
        CONNECTION_STATE_CLOSE_PIPE,
        CONNECTION_STATE_OPENED,
        CONNECTION_STATE_CLOSE_RCVD,
        CONNECTION_STATE_CLOSE_SENT,
        CONNECTION_STATE_DISCARDING,
        CONNECTION_STATE_END,
        CONNECTION_STATE_ERROR

    ctypedef CONNECTION_STATE_TAG CONNECTION_STATE

    ctypedef void (*ON_CONNECTION_STATE_CHANGED)(void* context, CONNECTION_STATE new_connection_state, CONNECTION_STATE previous_connection_state)
    ctypedef bint (*ON_NEW_ENDPOINT)(void* context, ENDPOINT_HANDLE new_endpoint)
    ctypedef void (*ON_CONNECTION_CLOSE_RECEIVED)(void* context, c_amqp_definitions.ERROR_HANDLE error)

    CONNECTION_HANDLE connection_create(c_xio.XIO_HANDLE io, const char* hostname, const char* container_id, ON_NEW_ENDPOINT on_new_endpoint, void* callback_context)
    CONNECTION_HANDLE connection_create2(c_xio.XIO_HANDLE xio, const char* hostname, const char* container_id, ON_NEW_ENDPOINT on_new_endpoint, void* callback_context, ON_CONNECTION_STATE_CHANGED on_connection_state_changed, void* on_connection_state_changed_context, c_xio.ON_IO_ERROR on_io_error, void* on_io_error_context)
    void connection_destroy(CONNECTION_HANDLE connection)
    int connection_open(CONNECTION_HANDLE connection)
    int connection_listen(CONNECTION_HANDLE connection)
    int connection_close(CONNECTION_HANDLE connection, const char* condition_value, const char* description, c_amqpvalue.AMQP_VALUE info)
    int connection_set_max_frame_size(CONNECTION_HANDLE connection, stdint.uint32_t max_frame_size)
    int connection_get_max_frame_size(CONNECTION_HANDLE connection, stdint.uint32_t* max_frame_size)
    int connection_set_channel_max(CONNECTION_HANDLE connection, stdint.uint16_t channel_max)
    int connection_get_channel_max(CONNECTION_HANDLE connection, stdint.uint16_t* channel_max)
    int connection_set_idle_timeout(CONNECTION_HANDLE connection, c_amqp_definitions.milliseconds idle_timeout)
    int connection_get_idle_timeout(CONNECTION_HANDLE connection, c_amqp_definitions.milliseconds* idle_timeout)
    int connection_set_properties(CONNECTION_HANDLE connection, c_amqp_definitions.fields properties)
    int connection_get_properties(CONNECTION_HANDLE connection, c_amqp_definitions.fields* properties)
    int connection_get_remote_max_frame_size(CONNECTION_HANDLE connection, stdint.uint32_t* remote_max_frame_size)
    int connection_set_remote_idle_timeout_empty_frame_send_ratio(CONNECTION_HANDLE connection, double idle_timeout_empty_frame_send_ratio)
    stdint.uint64_t connection_handle_deadlines(CONNECTION_HANDLE connection)
    void connection_dowork(CONNECTION_HANDLE connection) nogil
    void connection_set_trace(CONNECTION_HANDLE connection, bint trace_on)

    ON_CONNECTION_CLOSED_EVENT_SUBSCRIPTION_HANDLE connection_subscribe_on_connection_close_received(CONNECTION_HANDLE connection, ON_CONNECTION_CLOSE_RECEIVED on_connection_close_received, void* context)
    void connection_unsubscribe_on_connection_close_received(ON_CONNECTION_CLOSED_EVENT_SUBSCRIPTION_HANDLE event_subscription)




#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

import sys
from uamqp.utils import get_running_loop

def get_dict_with_loop_if_needed(loop):
    if sys.version_info >= (3, 10) and loop:
        raise ValueError("Starting Python 3.10, asyncio no longer supports loop as a parameter.")
    return {'loop': loop or get_running_loop()} if sys.version_info < (3, 10) else {}



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

from libc cimport stdint

cimport c_strings
cimport c_xio


cdef extern from "azure_c_shared_utility/platform.h":

    cdef enum PLATFORM_INFO_OPTION_TAG:
        PLATFORM_INFO_OPTION_DEFAULT,
        PLATFORM_INFO_OPTION_RETRIEVE_SQM,

    int platform_init()
    void platform_deinit()
    const c_xio.IO_INTERFACE_DESCRIPTION* platform_get_default_tlsio()
    c_strings.STRING_HANDLE platform_get_platform_info(PLATFORM_INFO_OPTION_TAG options)



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

# Python imports
from enum import Enum
import logging

# C imports
from libc cimport stdint

cimport c_message
cimport c_amqp_definitions
cimport c_amqpvalue


_logger = logging.getLogger(__name__)


class MessageBodyType(Enum):
    NoneType = c_message.MESSAGE_BODY_TYPE_TAG.MESSAGE_BODY_TYPE_NONE
    DataType = c_message.MESSAGE_BODY_TYPE_TAG.MESSAGE_BODY_TYPE_DATA
    SequenceType = c_message.MESSAGE_BODY_TYPE_TAG.MESSAGE_BODY_TYPE_SEQUENCE
    ValueType = c_message.MESSAGE_BODY_TYPE_TAG.MESSAGE_BODY_TYPE_VALUE


cdef message_factory(c_message.MESSAGE_HANDLE value):
    new_message = cMessage()
    new_message.wrap(value)
    return new_message


cpdef create_message():
    new_message = cMessage()
    new_message.create()
    return new_message


cdef class cMessage(StructBase):

    cdef c_message.MESSAGE_HANDLE _c_value

    def __cinit__(self):
        pass

    def __dealloc__(self):
        _logger.debug("Deallocating cMessage")
        self.destroy()

    cdef _create(self):
        if <void*>self._c_value is NULL:
            self._memory_error()

    cpdef destroy(self):
        try:
            if <void*>self._c_value is not NULL:
                _logger.debug("Destroying cMessage")
                c_message.message_destroy(self._c_value)
        except KeyboardInterrupt:
            pass
        finally:
            self._c_value = <c_message.MESSAGE_HANDLE>NULL

    cdef wrap(self, c_message.MESSAGE_HANDLE value):
        self.destroy()
        self._c_value = value
        self._create()

    cdef create(self):
        self.destroy()
        self._c_value = c_message.message_create()
        self._create()

    cpdef clone(self):
        cdef c_message.MESSAGE_HANDLE value
        value = c_message.message_clone(self._c_value)
        if <void*>value == NULL:
            self._value_error()
        return message_factory(value)

    @property
    def header(self):
        cdef c_amqp_definitions.HEADER_HANDLE value
        if c_message.message_get_header(self._c_value, &value) == 0:
            if <void*>value == NULL:
                return None
            new_obj = cHeader()
            new_obj.wrap(value)
            return new_obj
        else:
            self._value_error()

    @header.setter
    def header(self, cHeader value):
        if value is None:
            if c_message.message_set_header(
                self._c_value, <c_amqp_definitions.HEADER_HANDLE>NULL) != 0:
                    self._value_error()
        elif c_message.message_set_header(
            self._c_value, <c_amqp_definitions.HEADER_HANDLE>value._c_value) != 0:
                self._value_error()

    @property
    def delivery_annotations(self):
        cdef c_amqp_definitions.delivery_annotations value
        if c_message.message_get_delivery_annotations(self._c_value, &value) == 0:
            if <void*>value == NULL:
                return None
            new_obj = cDeliveryAnnotations()
            new_obj.wrap(value)
            return new_obj
        else:
            self._value_error()

    @delivery_annotations.setter
    def delivery_annotations(self, cDeliveryAnnotations value):
        if value is None:
            if c_message.message_set_delivery_annotations(
                self._c_value, <c_amqp_definitions.delivery_annotations>NULL) != 0:
                    self._value_error()
        elif c_message.message_set_delivery_annotations(
            self._c_value, <c_amqp_definitions.delivery_annotations>value._c_value) != 0:
                self._value_error()

    @property
    def message_annotations(self):
        cdef c_amqp_definitions.message_annotations value
        if c_message.message_get_message_annotations(self._c_value, &value) == 0:
            if <void*>value == NULL:
                return None
            new_obj = cMessageAnnotations()
            new_obj.wrap(value)
            return new_obj
        else:
            self._value_error()

    @message_annotations.setter
    def message_annotations(self, cMessageAnnotations value):
        if value is None:
            if c_message.message_set_message_annotations(
                self._c_value, <c_amqp_definitions.message_annotations>NULL) != 0:
                    self._value_error()
        elif c_message.message_set_message_annotations(
            self._c_value, <c_amqp_definitions.message_annotations>value._c_value) != 0:
                self._value_error()

    @property
    def properties(self):
        cdef c_amqp_definitions.PROPERTIES_HANDLE value
        if c_message.message_get_properties(self._c_value, &value) == 0:
            if <void*>value == NULL:
                return None
            new_obj = cProperties()
            new_obj.wrap(value)
            return new_obj
        else:
            self._value_error()

    @properties.setter
    def properties(self, cProperties value):
        if value is None:
            if c_message.message_set_properties(
                self._c_value, <c_amqp_definitions.PROPERTIES_HANDLE>NULL) != 0:
                    self._value_error()
        elif c_message.message_set_properties(
            self._c_value, <c_amqp_definitions.PROPERTIES_HANDLE>value._c_value) != 0:
                self._value_error()

    @property
    def application_properties(self):
        cdef c_amqpvalue.AMQP_VALUE value
        if c_message.message_get_application_properties(self._c_value, &value) == 0:
            if <void*>value == NULL:
                return None
            new_obj = cApplicationProperties()
            new_obj.wrap(value)
            return new_obj
        else:
            self._value_error()

    @application_properties.setter
    def application_properties(self, AMQPValue value):
        if value is None:
            if c_message.message_set_application_properties(
                self._c_value, <c_amqpvalue.AMQP_VALUE>NULL) != 0:
                    self._value_error()
        elif c_message.message_set_application_properties(
            self._c_value, <c_amqpvalue.AMQP_VALUE>value._c_value) != 0:
                self._value_error()

    @property
    def footer(self):
        cdef c_amqp_definitions.annotations value
        if c_message.message_get_footer(self._c_value, &value) == 0:
            if <void*>value == NULL:
                return None
            new_obj = cFooter()
            new_obj.wrap(value)
            return new_obj
        else:
            self._value_error()

    @footer.setter
    def footer(self, cFooter value):
        if value is None:
            if c_message.message_set_footer(
                self._c_value, <c_amqp_definitions.annotations>NULL) != 0:
                    self._value_error()
        elif c_message.message_set_footer(
            self._c_value, <c_amqp_definitions.annotations>value._c_value) != 0:
                self._value_error()

    @property
    def body_type(self):
        cdef c_message.MESSAGE_BODY_TYPE_TAG value
        if c_message.message_get_body_type(self._c_value, &value) == 0:
            return MessageBodyType(value)
        else:
            self._value_error()

    @property
    def message_format(self):
        cdef stdint.uint32_t value
        if c_message.message_get_message_format(self._c_value, &value) == 0:
            return value
        else:
            self._value_error()

    @message_format.setter
    def message_format(self, stdint.uint32_t value):
        if c_message.message_set_message_format(self._c_value, value) != 0:
                self._value_error()

    @property
    def delivery_tag(self):
        cdef c_amqpvalue.AMQP_VALUE value
        if c_message.message_get_delivery_tag(self._c_value, &value) == 0:
            if <void*>value == NULL:
                return None
            return value_factory(value)
        else:
            self._value_error()

    cpdef add_body_data(self, bytes value):
        cdef c_message.BINARY_DATA _binary
        length = len(value)
        bytes = value[:length]
        _binary.length = length
        _binary.bytes = bytes
        if c_message.message_add_body_amqp_data(self._c_value, _binary) != 0:
            self._value_error()

    cpdef get_body_data(self, size_t index):
        cdef c_message.BINARY_DATA _value
        if c_message.message_get_body_amqp_data_in_place(self._c_value, index, &_value) == 0:
            return _value.bytes[:_value.length]
        else:
            self._value_error()

    cpdef count_body_data(self):
        cdef size_t body_count
        if c_message.message_get_body_amqp_data_count(self._c_value, &body_count) == 0:
            return body_count
        else:
            self._value_error()

    cpdef set_body_value(self, AMQPValue value):
        if c_message.message_set_body_amqp_value(self._c_value, <c_amqpvalue.AMQP_VALUE>value._c_value) != 0:
            self._value_error()

    cpdef get_body_value(self):
        cdef c_amqpvalue.AMQP_VALUE _value
        cdef c_amqpvalue.AMQP_VALUE cloned
        if c_message.message_get_body_amqp_value_in_place(self._c_value, &_value) != 0:
            self._value_error()
        cloned = c_amqpvalue.amqpvalue_clone(_value)
        if <void*>cloned == NULL:
            self._value_error()
        return value_factory(cloned)

    cpdef add_body_sequence(self, AMQPValue sequence):
        if c_message.message_add_body_amqp_sequence(self._c_value, <c_amqpvalue.AMQP_VALUE>sequence._c_value) != 0:
            self._value_error()

    cpdef get_body_sequence(self, size_t index):
        cdef c_amqpvalue.AMQP_VALUE _value
        cdef c_amqpvalue.AMQP_VALUE cloned
        if c_message.message_get_body_amqp_sequence_in_place(self._c_value, index, &_value) != 0:
            self._value_error()
        cloned = c_amqpvalue.amqpvalue_clone(_value)
        if <void*>cloned == NULL:
            self._value_error()
        return value_factory(cloned)

    cpdef count_body_sequence(self):
        cdef size_t body_count
        if c_message.message_get_body_amqp_sequence_count(self._c_value, &body_count) == 0:
            return body_count
        else:
            self._value_error()


cdef class Messaging(object):

    @staticmethod
    def create_source(const char* address):
        cdef c_amqpvalue.AMQP_VALUE _value
        _value = c_message.messaging_create_source(address)
        if <void*>_value == NULL:
            raise MemoryError("Failed to allocate memory for messaging source.")
        return value_factory(_value)

    @staticmethod
    def create_target(const char* address):
        cdef c_amqpvalue.AMQP_VALUE _value
        _value = c_message.messaging_create_target(address)
        if <void*>_value == NULL:
            raise MemoryError("Failed to allocate memory for messaging target.")
        return value_factory(_value)

    @staticmethod
    def delivery_received(stdint.uint32_t section_number, stdint.uint64_t section_offset):
        cdef c_amqpvalue.AMQP_VALUE _value
        _value = c_message.messaging_delivery_received(section_number, section_offset)
        if <void*>_value == NULL:
            raise MemoryError("Failed to allocate memory for received delivery.")
        return value_factory(_value)

    @staticmethod
    def delivery_accepted():
        cdef c_amqpvalue.AMQP_VALUE _value
        _value = c_message.messaging_delivery_accepted()
        if <void*>_value == NULL:
            raise MemoryError("Failed to allocate memory for accepted delivery.")
        return value_factory(_value)

    @staticmethod
    def delivery_rejected(const char* error_condition, const char* error_description, cFields error_info=None):
        cdef c_amqpvalue.AMQP_VALUE _value

        if error_info is not None:
            delivery_fields = <c_amqp_definitions.fields>error_info._c_value
        else:
            delivery_fields = <c_amqp_definitions.fields>NULL

        _value = c_message.messaging_delivery_rejected(error_condition, error_description, delivery_fields)
        if <void*>_value == NULL:
            raise MemoryError("Failed to allocate memory for rejected delivery.")
        return value_factory(_value)

    @staticmethod
    def delivery_released():
        cdef c_amqpvalue.AMQP_VALUE _value
        _value = c_message.messaging_delivery_released()
        if <void*>_value == NULL:
            raise MemoryError("Failed to allocate memory for released delivery.")
        return value_factory(_value)

    @staticmethod
    def delivery_modified(bint delivery_failed, bint undeliverable_here, cFields message_annotations):
        _logger.debug("delivery modified: %r %r", delivery_failed, undeliverable_here)
        cdef c_amqpvalue.AMQP_VALUE _value
        _value = c_message.messaging_delivery_modified(delivery_failed, undeliverable_here, <c_amqp_definitions.fields>message_annotations._c_value)
        if <void*>_value == NULL:
            raise MemoryError("Failed to allocate memory for modified delivery.")
        return value_factory(_value)


cdef void destroy_amqp_objects_in_get_encoded_message_size(c_amqp_definitions.HEADER_HANDLE header,
    c_amqpvalue.AMQP_VALUE header_amqp_value, c_amqpvalue.AMQP_VALUE msg_annotations,
    c_amqpvalue.AMQP_VALUE footer, c_amqpvalue.AMQP_VALUE delivery_annotations,
    c_amqp_definitions.PROPERTIES_HANDLE properties, c_amqpvalue.AMQP_VALUE properties_amqp_value,
    c_amqpvalue.AMQP_VALUE application_properties, c_amqpvalue.AMQP_VALUE application_properties_value,
    c_amqpvalue.AMQP_VALUE body_amqp_value):

    if <void*>header != NULL:
        c_amqp_definitions.header_destroy(header)

    if <void*>header_amqp_value != NULL:
        c_amqpvalue.amqpvalue_destroy(header_amqp_value)

    if <void*>msg_annotations != NULL:
        c_amqpvalue.amqpvalue_destroy(msg_annotations)

    if <void*>footer != NULL:
        c_amqpvalue.amqpvalue_destroy(footer)

    if <void*>delivery_annotations != NULL:
        c_amqpvalue.amqpvalue_destroy(delivery_annotations)

    if <void*>properties != NULL:
        c_amqp_definitions.properties_destroy(properties)

    if <void*>properties_amqp_value != NULL:
        c_amqpvalue.amqpvalue_destroy(properties_amqp_value)

    if <void*>application_properties != NULL:
        c_amqpvalue.amqpvalue_destroy(application_properties)

    if <void*>application_properties_value != NULL:
        c_amqpvalue.amqpvalue_destroy(application_properties_value)

    if <void*>body_amqp_value != NULL:
        c_amqpvalue.amqpvalue_destroy(body_amqp_value)


cpdef size_t get_encoded_message_size(cMessage message, encoded_data):

        cdef c_message.MESSAGE_HANDLE c_msg
        c_msg = <c_message.MESSAGE_HANDLE>message._c_value

        cdef c_message.MESSAGE_BODY_TYPE_TAG message_body_type
        cdef c_amqp_definitions.message_format message_format
        cdef c_amqp_definitions.HEADER_HANDLE header = <c_amqp_definitions.HEADER_HANDLE>NULL
        cdef c_amqpvalue.AMQP_VALUE header_amqp_value = <c_amqpvalue.AMQP_VALUE>NULL
        cdef c_amqp_definitions.PROPERTIES_HANDLE properties = <c_amqp_definitions.PROPERTIES_HANDLE>NULL
        cdef c_amqpvalue.AMQP_VALUE properties_amqp_value = <c_amqpvalue.AMQP_VALUE>NULL
        cdef c_amqpvalue.AMQP_VALUE application_properties = <c_amqpvalue.AMQP_VALUE>NULL
        cdef c_amqpvalue.AMQP_VALUE application_properties_value = <c_amqpvalue.AMQP_VALUE>NULL
        cdef c_amqpvalue.AMQP_VALUE body_amqp_value = <c_amqpvalue.AMQP_VALUE>NULL
        cdef c_amqpvalue.AMQP_VALUE msg_annotations = <c_amqpvalue.AMQP_VALUE>NULL
        cdef c_amqpvalue.AMQP_VALUE footer = <c_amqpvalue.AMQP_VALUE>NULL
        cdef c_amqpvalue.AMQP_VALUE delivery_annotations = <c_amqpvalue.AMQP_VALUE>NULL
        cdef c_amqpvalue.AMQP_VALUE message_body_amqp_value
        cdef c_amqpvalue.AMQP_VALUE message_body_amqp_sequence
        cdef c_message.BINARY_DATA binary_data
        cdef c_amqpvalue.AMQP_VALUE body_amqp_data
        cdef c_amqpvalue.AMQP_VALUE body_amqp_sequence
        cdef c_amqpvalue.amqp_binary binary_value
        cdef size_t encoded_size
        cdef size_t total_encoded_size = 0
        cdef size_t body_data_count = 0
        cdef size_t body_sequence_count = 0

        # message format
        if (c_message.message_get_body_type(c_msg, &message_body_type) != 0) or (c_message.message_get_message_format(c_msg, &message_format) != 0):
            raise ValueError("Failure getting message body type and/or message format")

        # message header
        if c_message.message_get_header(c_msg, &header) == 0 and <void*>header != NULL:
            header_amqp_value = c_amqp_definitions.amqpvalue_create_header(header)
            if <void*>header_amqp_value == NULL:
                _logger.debug("Cannot create header AMQP value")
                destroy_amqp_objects_in_get_encoded_message_size(header, header_amqp_value, msg_annotations, footer, delivery_annotations,
                    properties, properties_amqp_value, application_properties, application_properties_value,
                    body_amqp_value)
                raise MemoryError("Cannot get cMessage header.")
            else:
                if c_amqpvalue.amqpvalue_get_encoded_size(header_amqp_value, &encoded_size) != 0:
                    _logger.debug("Cannot obtain header encoded size")
                    destroy_amqp_objects_in_get_encoded_message_size(header, header_amqp_value, msg_annotations, footer, delivery_annotations,
                        properties, properties_amqp_value, application_properties, application_properties_value,
                        body_amqp_value)
                    raise ValueError("Cannot obtain header encoded size")
                else:
                    total_encoded_size += encoded_size

        # message annotations
        if c_message.message_get_message_annotations(c_msg, &msg_annotations) == 0 and <void*>msg_annotations != NULL:
            if c_amqpvalue.amqpvalue_get_encoded_size(msg_annotations, &encoded_size) != 0:
                _logger.debug("Cannot obtain message annotations encoded size")
                destroy_amqp_objects_in_get_encoded_message_size(header, header_amqp_value, msg_annotations, footer, delivery_annotations,
                    properties, properties_amqp_value, application_properties, application_properties_value,
                    body_amqp_value)
                raise ValueError("Cannot obtain message annotations encoded size")
            else:
                total_encoded_size += encoded_size

        # properties
        if c_message.message_get_properties(c_msg, &properties) == 0 and <void*>properties != NULL:
            properties_amqp_value = c_amqp_definitions.amqpvalue_create_properties(properties)
            if <void*>properties_amqp_value == NULL:
                _logger.debug("Cannot create properties AMQP value")
                destroy_amqp_objects_in_get_encoded_message_size(header, header_amqp_value, msg_annotations, footer, delivery_annotations,
                    properties, properties_amqp_value, application_properties, application_properties_value,
                    body_amqp_value)
                raise MemoryError("Cannot get cMessage properties.")
            else:
                if c_amqpvalue.amqpvalue_get_encoded_size(properties_amqp_value, &encoded_size) != 0:
                    _logger.debug("Cannot obtain message properties encoded size")
                    destroy_amqp_objects_in_get_encoded_message_size(header, header_amqp_value, msg_annotations, footer, delivery_annotations,
                        properties, properties_amqp_value, application_properties, application_properties_value,
                        body_amqp_value)
                    raise ValueError("Cannot obtain message properties encoded size")
                else:
                    total_encoded_size += encoded_size

        # application properties
        if c_message.message_get_application_properties(c_msg, &application_properties) == 0 and <void*>application_properties != NULL:
            application_properties_value = c_amqp_definitions.amqpvalue_create_application_properties(application_properties)
            if <void*>application_properties_value == NULL:
                _logger.debug("Cannot create application properties AMQP value")
                destroy_amqp_objects_in_get_encoded_message_size(header, header_amqp_value, msg_annotations, footer, delivery_annotations,
                    properties, properties_amqp_value, application_properties, application_properties_value,
                    body_amqp_value)
                raise MemoryError("Cannot get cMessage application properties.")
            else:
                if c_amqpvalue.amqpvalue_get_encoded_size(application_properties_value, &encoded_size) != 0:
                    _logger.debug("Cannot obtain application properties encoded size")
                    destroy_amqp_objects_in_get_encoded_message_size(header, header_amqp_value, msg_annotations, footer, delivery_annotations,
                        properties, properties_amqp_value, application_properties, application_properties_value,
                        body_amqp_value)
                    raise ValueError("Cannot obtain application properties encoded size")
                else:
                    total_encoded_size += encoded_size

        # footer
        if c_message.message_get_footer(c_msg, &footer) == 0 and <void*>footer != NULL:
            if c_amqpvalue.amqpvalue_get_encoded_size(footer, &encoded_size) != 0:
                _logger.debug("Cannot obtain footer encoded size")
                destroy_amqp_objects_in_get_encoded_message_size(header, header_amqp_value, msg_annotations, footer, delivery_annotations,
                    properties, properties_amqp_value, application_properties, application_properties_value,
                    body_amqp_value)
                raise ValueError("Cannot obtain footer encoded size")
            else:
                total_encoded_size += encoded_size

        # delivery annotations
        if c_message.message_get_delivery_annotations(c_msg, &delivery_annotations) == 0 and <void*>delivery_annotations != NULL:
            if c_amqpvalue.amqpvalue_get_encoded_size(delivery_annotations, &encoded_size) != 0:
                _logger.debug("Cannot obtain delivery annotations encoded size")
                destroy_amqp_objects_in_get_encoded_message_size(header, header_amqp_value, msg_annotations, footer, delivery_annotations,
                    properties, properties_amqp_value, application_properties, application_properties_value,
                    body_amqp_value)
                raise ValueError("Cannot obtain delivery annotations encoded size")
            else:
                total_encoded_size += encoded_size

        # value body
        if message_body_type == c_message.MESSAGE_BODY_TYPE_TAG.MESSAGE_BODY_TYPE_VALUE:
            if c_message.message_get_body_amqp_value_in_place(c_msg, &message_body_amqp_value) != 0:
                destroy_amqp_objects_in_get_encoded_message_size(header, header_amqp_value, msg_annotations, footer, delivery_annotations,
                    properties, properties_amqp_value, application_properties, application_properties_value,
                    body_amqp_value)
                raise ValueError("Cannot obtain AMQP value from body")
            body_amqp_value = c_amqp_definitions.amqpvalue_create_amqp_value(message_body_amqp_value)
            if <void*>body_amqp_value == NULL:
                destroy_amqp_objects_in_get_encoded_message_size(header, header_amqp_value, msg_annotations, footer, delivery_annotations,
                    properties, properties_amqp_value, application_properties, application_properties_value,
                    body_amqp_value)
                raise MemoryError("Cannot create body AMQP value")
            else:
                if c_amqpvalue.amqpvalue_get_encoded_size(body_amqp_value, &encoded_size) != 0:
                    _logger.debug("Cannot get body AMQP value encoded size")
                    destroy_amqp_objects_in_get_encoded_message_size(header, header_amqp_value, msg_annotations, footer, delivery_annotations,
                        properties, properties_amqp_value, application_properties, application_properties_value,
                        body_amqp_value)
                    raise ValueError("Cannot get body AMQP value encoded size")
                else:
                    total_encoded_size += encoded_size

        # data body
        if message_body_type == c_message.MESSAGE_BODY_TYPE_TAG.MESSAGE_BODY_TYPE_DATA:
            if c_message.message_get_body_amqp_data_count(c_msg, &body_data_count) != 0:
                destroy_amqp_objects_in_get_encoded_message_size(header, header_amqp_value, msg_annotations, footer, delivery_annotations,
                    properties, properties_amqp_value, application_properties, application_properties_value,
                    body_amqp_value)
                raise ValueError("Cannot get body AMQP data count")
            if body_data_count == 0:
                destroy_amqp_objects_in_get_encoded_message_size(header, header_amqp_value, msg_annotations, footer, delivery_annotations,
                    properties, properties_amqp_value, application_properties, application_properties_value,
                    body_amqp_value)
                raise ValueError("Body data count is zero")
            for i in range(body_data_count):
                if c_message.message_get_body_amqp_data_in_place(c_msg, i, &binary_data) != 0:
                    destroy_amqp_objects_in_get_encoded_message_size(header, header_amqp_value, msg_annotations, footer, delivery_annotations,
                        properties, properties_amqp_value, application_properties, application_properties_value,
                        body_amqp_value)
                    raise ValueError("Cannot get body AMQP data {}".format(i))

                binary_value.bytes = binary_data.bytes
                binary_value.length = <stdint.uint32_t>binary_data.length
                body_amqp_data = c_amqp_definitions.amqpvalue_create_data(binary_value)
                if <void*>body_amqp_data == NULL:
                    destroy_amqp_objects_in_get_encoded_message_size(header, header_amqp_value, msg_annotations, footer, delivery_annotations,
                        properties, properties_amqp_value, application_properties, application_properties_value,
                        body_amqp_value)
                    raise MemoryError("Cannot create body AMQP data")
                else:
                    if c_amqpvalue.amqpvalue_get_encoded_size(body_amqp_data, &encoded_size) != 0:
                        _logger.debug("Cannot get body AMQP data encoded size")
                        destroy_amqp_objects_in_get_encoded_message_size(header, header_amqp_value, msg_annotations, footer, delivery_annotations,
                            properties, properties_amqp_value, application_properties, application_properties_value,
                            body_amqp_value)
                        c_amqpvalue.amqpvalue_destroy(body_amqp_data)
                        raise ValueError("Cannot get body AMQP data encoded size")
                    else:
                        total_encoded_size += encoded_size
                    c_amqpvalue.amqpvalue_destroy(body_amqp_data)

        # sequence body
        if message_body_type == c_message.MESSAGE_BODY_TYPE_TAG.MESSAGE_BODY_TYPE_SEQUENCE:
            if c_message.message_get_body_amqp_sequence_count(c_msg, &body_sequence_count) != 0:
                destroy_amqp_objects_in_get_encoded_message_size(header, header_amqp_value, msg_annotations, footer, delivery_annotations,
                    properties, properties_amqp_value, application_properties, application_properties_value,
                    body_amqp_value)
                raise ValueError("Cannot get body AMQP sequence count")
            if body_sequence_count == 0:
                destroy_amqp_objects_in_get_encoded_message_size(header, header_amqp_value, msg_annotations, footer, delivery_annotations,
                    properties, properties_amqp_value, application_properties, application_properties_value,
                    body_amqp_value)
                raise ValueError("Body sequence count is zero")
            for i in range(body_sequence_count):
                if c_message.message_get_body_amqp_sequence_in_place(c_msg, i, &message_body_amqp_sequence) != 0:
                    destroy_amqp_objects_in_get_encoded_message_size(header, header_amqp_value, msg_annotations, footer, delivery_annotations,
                        properties, properties_amqp_value, application_properties, application_properties_value,
                        body_amqp_value)
                    raise ValueError("Cannot get body AMQP sequence {}".format(i))

                body_amqp_sequence = c_amqp_definitions.amqpvalue_create_amqp_sequence(message_body_amqp_sequence);
                if <void*> body_amqp_sequence == NULL:
                    destroy_amqp_objects_in_get_encoded_message_size(header, header_amqp_value, msg_annotations, footer, delivery_annotations,
                        properties, properties_amqp_value, application_properties, application_properties_value,
                        body_amqp_value)
                    raise MemoryError("Cannot create body AMQP sequence")
                else:
                    if c_amqpvalue.amqpvalue_get_encoded_size(body_amqp_sequence, &encoded_size) != 0:
                        _logger.debug("Cannot get body AMQP sequence encoded size")
                        destroy_amqp_objects_in_get_encoded_message_size(header, header_amqp_value, msg_annotations, footer, delivery_annotations,
                            properties, properties_amqp_value, application_properties, application_properties_value,
                            body_amqp_value)
                        c_amqpvalue.amqpvalue_destroy(body_amqp_sequence)
                        raise ValueError("Cannot get body AMQP sequence encoded size")
                    else:
                        total_encoded_size += encoded_size
                    c_amqpvalue.amqpvalue_destroy(body_amqp_sequence)

        # encode
        if <void*>header != NULL:
            if c_amqpvalue.amqpvalue_encode(
                    header_amqp_value,
                    <c_amqpvalue.AMQPVALUE_ENCODER_OUTPUT>encode_bytes_callback,
                    <void*>encoded_data) != 0:
                destroy_amqp_objects_in_get_encoded_message_size(header, header_amqp_value, msg_annotations, footer, delivery_annotations,
                    properties, properties_amqp_value, application_properties, application_properties_value,
                    body_amqp_value)
                raise ValueError("Cannot encode header value")
        if <void*>msg_annotations != NULL:
            if c_amqpvalue.amqpvalue_encode(
                    msg_annotations,
                    <c_amqpvalue.AMQPVALUE_ENCODER_OUTPUT>encode_bytes_callback,
                    <void*>encoded_data) != 0:
                destroy_amqp_objects_in_get_encoded_message_size(header, header_amqp_value, msg_annotations, footer, delivery_annotations,
                    properties, properties_amqp_value, application_properties, application_properties_value,
                    body_amqp_value)
                raise ValueError("Cannot encode message annotations value")
        if <void*>properties != NULL:
            if c_amqpvalue.amqpvalue_encode(
                    properties_amqp_value,
                    <c_amqpvalue.AMQPVALUE_ENCODER_OUTPUT>encode_bytes_callback,
                    <void*>encoded_data) != 0:
                destroy_amqp_objects_in_get_encoded_message_size(header, header_amqp_value, msg_annotations, footer, delivery_annotations,
                    properties, properties_amqp_value, application_properties, application_properties_value,
                    body_amqp_value)
                raise ValueError("Cannot encode message properties value")
        if <void*>application_properties != NULL:
            if c_amqpvalue.amqpvalue_encode(
                    application_properties_value,
                    <c_amqpvalue.AMQPVALUE_ENCODER_OUTPUT>encode_bytes_callback,
                    <void*>encoded_data) != 0:
                destroy_amqp_objects_in_get_encoded_message_size(header, header_amqp_value, msg_annotations, footer, delivery_annotations,
                    properties, properties_amqp_value, application_properties, application_properties_value,
                    body_amqp_value)
                raise ValueError("Cannot encode application properties value")
        if <void*>footer != NULL:
            if c_amqpvalue.amqpvalue_encode(
                    footer,
                    <c_amqpvalue.AMQPVALUE_ENCODER_OUTPUT>encode_bytes_callback,
                    <void*>encoded_data) != 0:
                destroy_amqp_objects_in_get_encoded_message_size(header, header_amqp_value, msg_annotations, footer, delivery_annotations,
                    properties, properties_amqp_value, application_properties, application_properties_value,
                    body_amqp_value)
                raise ValueError("Cannot encode footer value")
        if <void*>delivery_annotations != NULL:
            if c_amqpvalue.amqpvalue_encode(
                    delivery_annotations,
                    <c_amqpvalue.AMQPVALUE_ENCODER_OUTPUT>encode_bytes_callback,
                    <void*>encoded_data) != 0:
                destroy_amqp_objects_in_get_encoded_message_size(header, header_amqp_value, msg_annotations, footer, delivery_annotations,
                    properties, properties_amqp_value, application_properties, application_properties_value,
                    body_amqp_value)
                raise ValueError("Cannot encode delivery annotations value")
        if message_body_type == c_message.MESSAGE_BODY_TYPE_TAG.MESSAGE_BODY_TYPE_VALUE:
            if c_amqpvalue.amqpvalue_encode(
                    body_amqp_value,
                    <c_amqpvalue.AMQPVALUE_ENCODER_OUTPUT>encode_bytes_callback,
                    <void*>encoded_data) != 0:
                destroy_amqp_objects_in_get_encoded_message_size(header, header_amqp_value, msg_annotations, footer, delivery_annotations,
                    properties, properties_amqp_value, application_properties, application_properties_value,
                    body_amqp_value)
                raise ValueError("Cannot encode body AMQP value")
        if message_body_type == c_message.MESSAGE_BODY_TYPE_TAG.MESSAGE_BODY_TYPE_DATA:
            for i in range(body_data_count):
                if c_message.message_get_body_amqp_data_in_place(c_msg, i, &binary_data) != 0:
                    destroy_amqp_objects_in_get_encoded_message_size(header, header_amqp_value, msg_annotations, footer, delivery_annotations,
                        properties, properties_amqp_value, application_properties, application_properties_value,
                        body_amqp_value)
                    raise ValueError("Cannot get body AMQP data {}".format(i))

                binary_value.bytes = binary_data.bytes
                binary_value.length = <stdint.uint32_t>binary_data.length
                body_amqp_data = c_amqp_definitions.amqpvalue_create_data(binary_value)
                if <void*>body_amqp_data == NULL:
                    destroy_amqp_objects_in_get_encoded_message_size(header, header_amqp_value, msg_annotations, footer, delivery_annotations,
                        properties, properties_amqp_value, application_properties, application_properties_value,
                        body_amqp_value)
                    raise MemoryError("Cannot create body AMQP data")
                else:
                    if c_amqpvalue.amqpvalue_encode(
                            body_amqp_data,
                            <c_amqpvalue.AMQPVALUE_ENCODER_OUTPUT>encode_bytes_callback,
                            <void*>encoded_data) != 0:
                        destroy_amqp_objects_in_get_encoded_message_size(header, header_amqp_value, msg_annotations, footer, delivery_annotations,
                            properties, properties_amqp_value, application_properties, application_properties_value,
                            body_amqp_value)
                        c_amqpvalue.amqpvalue_destroy(body_amqp_data)
                        raise ValueError("Cannot encode body AMQP value")
                    c_amqpvalue.amqpvalue_destroy(body_amqp_data)
        if message_body_type == c_message.MESSAGE_BODY_TYPE_TAG.MESSAGE_BODY_TYPE_SEQUENCE:
            for i in range(body_sequence_count):
                if c_message.message_get_body_amqp_sequence_in_place(c_msg, i, &message_body_amqp_sequence) != 0:
                    destroy_amqp_objects_in_get_encoded_message_size(header, header_amqp_value, msg_annotations, footer, delivery_annotations,
                        properties, properties_amqp_value, application_properties, application_properties_value,
                        body_amqp_value)
                    raise ValueError("Cannot get body AMQP sequence {}".format(i))

                body_amqp_sequence = c_amqp_definitions.amqpvalue_create_amqp_sequence(message_body_amqp_sequence);
                if <void*> body_amqp_sequence == NULL:
                    destroy_amqp_objects_in_get_encoded_message_size(header, header_amqp_value, msg_annotations, footer, delivery_annotations,
                        properties, properties_amqp_value, application_properties, application_properties_value,
                        body_amqp_value)
                    raise MemoryError("Cannot create body AMQP sequence")
                else:
                    if c_amqpvalue.amqpvalue_encode(
                            body_amqp_sequence,
                            <c_amqpvalue.AMQPVALUE_ENCODER_OUTPUT>encode_bytes_callback,
                            <void*>encoded_data) != 0:
                        destroy_amqp_objects_in_get_encoded_message_size(header, header_amqp_value, msg_annotations, footer, delivery_annotations,
                            properties, properties_amqp_value, application_properties, application_properties_value,
                            body_amqp_value)
                        c_amqpvalue.amqpvalue_destroy(body_amqp_sequence)
                        raise ValueError("Cannot get body AMQP sequence encoded size")
                    c_amqpvalue.amqpvalue_destroy(body_amqp_sequence)

        destroy_amqp_objects_in_get_encoded_message_size(header, header_amqp_value, msg_annotations, footer, delivery_annotations,
            properties, properties_amqp_value, application_properties, application_properties_value,
            body_amqp_value)

        return total_encoded_size


cdef class cMessageDecoder(object):

    cdef c_message.MESSAGE_HANDLE decoded_message
    cdef const char* decode_error

    def __cinit__(self):
        self.decoded_message = c_message.message_create()
        if <void*>self.decoded_message == NULL:
            raise MemoryError("Failed to create message decoder.")
        self.decode_error = <const char*>NULL

    def __dealloc__(self):
        _logger.debug("Deallocating cMessageDecoder")


cpdef cMessage decode_message(stdint.uint32_t payload_size, const unsigned char* payload_bytes):
    cdef c_amqpvalue.AMQPVALUE_DECODER_HANDLE amqpvalue_decoder
    cdef c_message.MESSAGE_HANDLE result
    message_decoder = cMessageDecoder()

    amqpvalue_decoder = c_amqpvalue.amqpvalue_decoder_create(<c_amqpvalue.ON_VALUE_DECODED>decode_message_data, <void*>message_decoder);
    if <void*>amqpvalue_decoder == NULL:
        raise MemoryError("Cannot create AMQP value decoder")
    else:
        if c_amqpvalue.amqpvalue_decode_bytes(amqpvalue_decoder, payload_bytes, payload_size) != 0:
            raise ValueError("Cannot decode bytes")
        else:
            if <void*>message_decoder.decode_error != NULL:
                raise ValueError(message_decoder.decode_error.decode('UTF-8'))
            else:
                result = message_decoder.decoded_message

        c_amqpvalue.amqpvalue_decoder_destroy(amqpvalue_decoder)
    return message_factory(result)


cdef void decode_message_data(void* context, c_amqpvalue.AMQP_VALUE decoded_value):
    message_decoder = <cMessageDecoder>context
    cdef c_message.MESSAGE_HANDLE decoded_message
    cdef c_amqpvalue.AMQP_VALUE descriptor
    cdef c_amqp_definitions.PROPERTIES_HANDLE properties
    cdef c_amqp_definitions.annotations delivery_annotations
    cdef c_amqp_definitions.annotations message_annotations
    cdef c_amqp_definitions.HEADER_HANDLE header
    cdef c_amqp_definitions.annotations footer
    cdef c_message.MESSAGE_BODY_TYPE_TAG body_type
    cdef c_amqpvalue.AMQP_VALUE body_amqp_value
    cdef c_amqpvalue.AMQP_VALUE body_data_value
    cdef c_amqp_definitions.data data_value
    cdef c_message.BINARY_DATA binary_data

    decoded_message = message_decoder.decoded_message
    descriptor = c_amqpvalue.amqpvalue_get_inplace_descriptor(decoded_value)

    if c_amqp_definitions.is_application_properties_type_by_descriptor(descriptor):
        if c_message.message_set_application_properties(decoded_message, decoded_value) != 0:
            message_decoder.decode_error = b"Error setting application properties on received message"

    elif c_amqp_definitions.is_properties_type_by_descriptor(descriptor):
        if c_amqp_definitions.amqpvalue_get_properties(decoded_value, &properties) != 0:
            message_decoder.decode_error = b"Error getting message properties"

        else:
            if c_message.message_set_properties(decoded_message, properties) != 0:
                message_decoder.decode_error = b"Error setting message properties on received message"
            c_amqp_definitions.properties_destroy(properties)

    elif c_amqp_definitions.is_delivery_annotations_type_by_descriptor(descriptor):
        delivery_annotations = c_amqpvalue.amqpvalue_get_inplace_described_value(decoded_value)
        if <void*>delivery_annotations == NULL:
            message_decoder.decode_error = b"Error getting delivery annotations"
        else:
            if c_message.message_set_delivery_annotations(decoded_message, delivery_annotations) != 0:
                message_decoder.decode_error = b"Error setting delivery annotations on received message"

    elif c_amqp_definitions.is_message_annotations_type_by_descriptor(descriptor):
        message_annotations = c_amqpvalue.amqpvalue_get_inplace_described_value(decoded_value)
        if <void*>message_annotations == NULL:
            message_decoder.decode_error = b"Error getting message annotations"
        else:
            if c_message.message_set_message_annotations(decoded_message, message_annotations) != 0:
                message_decoder.decode_error = b"Error setting message annotations on received message"

    elif c_amqp_definitions.is_header_type_by_descriptor(descriptor):
        if c_amqp_definitions.amqpvalue_get_header(decoded_value, &header) != 0:
            message_decoder.decode_error = b"Error getting message header"
        else:
            if c_message.message_set_header(decoded_message, header) != 0:
                message_decoder.decode_error = b"Error setting message header on received message"
            c_amqp_definitions.header_destroy(header)

    elif c_amqp_definitions.is_footer_type_by_descriptor(descriptor):
        footer = c_amqpvalue.amqpvalue_get_inplace_described_value(decoded_value);
        if <void*>footer == NULL:
            message_decoder.decode_error = b"Error getting message footer"
        else:
            if c_message.message_set_footer(decoded_message, footer) != 0:
                message_decoder.decode_error = b"Error setting message footer on received message"

    elif c_amqp_definitions.is_amqp_value_type_by_descriptor(descriptor):
        if c_message.message_get_body_type(decoded_message, &body_type) != 0:
            message_decoder.decode_error = b"Error getting message body type"

        else:
            if body_type != c_message.MESSAGE_BODY_TYPE_TAG.MESSAGE_BODY_TYPE_NONE:
                message_decoder.decode_error = b"Body already set on received message"

            else:
                body_amqp_value = c_amqpvalue.amqpvalue_get_inplace_described_value(decoded_value)
                if <void*>body_amqp_value == NULL:
                    message_decoder.decode_error = b"Error getting body AMQP value"
                else:
                    if c_message.message_set_body_amqp_value(decoded_message, body_amqp_value) != 0:
                        message_decoder.decode_error = b"Error setting body AMQP value on received message"

    elif c_amqp_definitions.is_data_type_by_descriptor(descriptor):
        if c_message.message_get_body_type(decoded_message, &body_type) != 0:
            message_decoder.decode_error = b"Error getting message body type"

        else:
            if (body_type != c_message.MESSAGE_BODY_TYPE_TAG.MESSAGE_BODY_TYPE_NONE) and (body_type != c_message.MESSAGE_BODY_TYPE_TAG.MESSAGE_BODY_TYPE_DATA):
                message_decoder.decode_error = b"Message body type already set to something different than AMQP DATA"

            else:
                body_data_value = c_amqpvalue.amqpvalue_get_inplace_described_value(decoded_value)
                if <void*>body_data_value == NULL:
                    message_decoder.decode_error = b"Error getting body DATA value"

                else:
                    if c_amqpvalue.amqpvalue_get_binary(body_data_value, &data_value) != 0:
                        message_decoder.decode_error = b"Error getting body DATA AMQP value"

                    else:
                        binary_data.bytes = <const unsigned char*>data_value.bytes
                        binary_data.length = data_value.length
                        if c_message.message_add_body_amqp_data(decoded_message, binary_data) != 0:
                            message_decoder.decode_error = b"Error adding body DATA to received message"



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

# Python imports
import logging

# C imports
cimport c_async_operation


_logger = logging.getLogger(__name__)


cdef class AsyncOperation(StructBase):

    cdef c_async_operation.ASYNC_OPERATION_HANDLE _c_value

    def __cinit__(self):
        pass

    cdef _create(self):
        if <void*>self._c_value is NULL:
            self._memory_error()

    cpdef destroy(self):
        if <void*>self._c_value is not NULL:
            _logger.debug("Destroying AsyncOperation")
            c_async_operation.async_operation_destroy(self._c_value)
            self._c_value = <c_async_operation.ASYNC_OPERATION_HANDLE>NULL

    cdef wrap(self, c_async_operation.ASYNC_OPERATION_HANDLE value):
        self.destroy()
        self._c_value = value
        self._create()

    cpdef cancel(self):
        if c_async_operation.async_operation_cancel(self._c_value) != 0:
            self._value_error()



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

import logging
import uuid

import six
from uamqp import c_uamqp, constants, errors, utils

_logger = logging.getLogger(__name__)


class MessageSender(object):
    """A Message Sender that opens its own exclsuive Link on an
    existing Session.

    :ivar send_settle_mode: The mode by which to settle message send
     operations. If set to `Unsettled`, the client will wait for a confirmation
     from the service that the message was successfully send. If set to 'Settled',
     the client will not wait for confirmation and assume success.
    :vartype send_settle_mode: ~uamqp.constants.SenderSettleMode
    :ivar receive_settle_mode: The mode by which to settle message receive
     operations. If set to `PeekLock`, the receiver will lock a message once received until
     the client accepts or rejects the message. If set to `ReceiveAndDelete`, the service
     will assume successful receipt of the message and clear it from the queue. The
     default is `PeekLock`.
    :vartype receive_settle_mode: ~uamqp.constants.ReceiverSettleMode
    :ivar max_message_size: The maximum allowed message size negotiated for the Link.
    :vartype max_message_size: int

    :param session: The underlying Session with which to send.
    :type session: ~uamqp.session.Session
    :param source: The name of source (i.e. the client).
    :type source: str or bytes
    :param target: The AMQP endpoint to send to.
    :type target: ~uamqp.address.Target
    :param name: A unique name for the sender. If not specified a GUID will be used.
    :type name: str or bytes
    :param send_settle_mode: The mode by which to settle message send
     operations. If set to `Unsettled`, the client will wait for a confirmation
     from the service that the message was successfully sent. If set to 'Settled',
     the client will not wait for confirmation and assume success.
    :type send_settle_mode: ~uamqp.constants.SenderSettleMode
    :param receive_settle_mode: The mode by which to settle message receive
     operations. If set to `PeekLock`, the receiver will lock a message once received until
     the client accepts or rejects the message. If set to `ReceiveAndDelete`, the service
     will assume successful receipt of the message and clear it from the queue. The
     default is `PeekLock`.
    :type receive_settle_mode: ~uamqp.constants.ReceiverSettleMode
    :param desired_capabilities: The extension capabilities desired from the peer endpoint.
     To create a desired_capabilities object, please do as follows:
        - 1. Create an array of desired capability symbols: `capabilities_symbol_array = [types.AMQPSymbol(string)]`
        - 2. Transform the array to AMQPValue object: `utils.data_factory(types.AMQPArray(capabilities_symbol_array))`
    :type desired_capabilities: ~uamqp.c_uamqp.AMQPValue
    :param max_message_size: The maximum allowed message size negotiated for the Link.
    :type max_message_size: int
    :param link_credit: The sender Link credit that determines how many
     messages the Link will attempt to handle per connection iteration.
    :type link_credit: int
    :param properties: Metadata to be sent in the Link ATTACH frame.
    :type properties: dict
    :param error_policy: A policy for parsing errors on link, connection and message
     disposition to determine whether the error should be retryable.
    :type error_policy: ~uamqp.errors.ErrorPolicy
    :param debug: Whether to turn on network trace logs. If `True`, trace logs
     will be logged at INFO level. Default is `False`.
    :type debug: bool
    :param encoding: The encoding to use for parameters supplied as strings.
     Default is 'UTF-8'
    :type encoding: str
    """

    def __init__(self, session, source, target,
                 name=None,
                 send_settle_mode=constants.SenderSettleMode.Unsettled,
                 receive_settle_mode=constants.ReceiverSettleMode.PeekLock,
                 max_message_size=constants.MAX_MESSAGE_LENGTH_BYTES,
                 link_credit=None,
                 properties=None,
                 error_policy=None,
                 debug=False,
                 encoding='UTF-8',
                 desired_capabilities=None):
        # pylint: disable=protected-access
        if name:
            self.name = name.encode(encoding) if isinstance(name, six.text_type) else name
        else:
            self.name = str(uuid.uuid4()).encode(encoding)
        source = source.encode(encoding) if isinstance(source, six.text_type) else source
        role = constants.Role.Sender

        self.source = c_uamqp.Messaging.create_source(source)
        self.target = target._address.value
        self.error_policy = error_policy or errors.ErrorPolicy()
        self._conn = session._conn
        self._session = session
        self._link = c_uamqp.create_link(session._session, self.name, role.value, self.source, self.target)
        self._link.max_message_size = max_message_size
        self._link.subscribe_to_detach_event(self)

        if link_credit:
            self._link.set_prefetch_count(link_credit)
        if properties:
            self._link.set_attach_properties(utils.data_factory(properties, encoding=encoding))
        if send_settle_mode:
            self.send_settle_mode = send_settle_mode
        if receive_settle_mode:
            self.receive_settle_mode = receive_settle_mode
        if max_message_size:
            self.max_message_size = max_message_size
        if desired_capabilities:
            self._link.set_desired_capabilities(desired_capabilities)

        self._sender = c_uamqp.create_message_sender(self._link, self)
        self._sender.set_trace(debug)
        self._state = constants.MessageSenderState.Idle
        self._error = None

    def __enter__(self):
        """Open the MessageSender in a context manager."""
        self.open()
        return self

    def __exit__(self, *args):
        """Close the MessageSender when exiting a context manager."""
        self.destroy()

    def _detach_received(self, error):
        """Callback called when a link DETACH frame is received.
        This callback will process the received DETACH error to determine if
        the link is recoverable or whether it should be shutdown.

        :param error: The error information from the detach
         frame.
        :type error: ~uamqp.errors.ErrorResponse
        """
        # pylint: disable=protected-access
        if error:
            condition = error.condition
            description = error.description
            info = error.info
        else:
            condition = b"amqp:unknown-error"
            description = None
            info = None
        self._error = errors._process_link_error(self.error_policy, condition, description, info)
        _logger.info("Received Link detach event: %r\nLink: %r\nDescription: %r"
                     "\nDetails: %r\nRetryable: %r\nConnection: %r",
                     condition, self.name, description, info, self._error.action.retry,
                     self._session._connection.container_id)

    def _state_changed(self, previous_state, new_state):
        """Callback called whenever the underlying Sender undergoes a change
        of state. This function wraps the states as Enums to prepare for
        calling the public callback.

        :param previous_state: The previous Sender state.
        :type previous_state: int
        :param new_state: The new Sender state.
        :type new_state: int
        """
        try:
            try:
                _previous_state = constants.MessageSenderState(previous_state)
            except ValueError:
                _previous_state = previous_state
            try:
                _new_state = constants.MessageSenderState(new_state)
            except ValueError:
                _new_state = new_state
            if _previous_state == constants.MessageSenderState.Opening \
                    and _new_state == constants.MessageSenderState.Error:
                _logger.info("Sender link failed to open - expecting to receive DETACH frame.")
            elif self._session._link_error:  # pylint: disable=protected-access
                _logger.info("Sender link ATTACH frame invalid - expecting to receive DETACH frame.")
            else:
                self.on_state_changed(_previous_state, _new_state)
        except KeyboardInterrupt:
            _logger.error("Received shutdown signal while updating sender state from {} to {}".format(
                previous_state, new_state))
            self._error = errors.AMQPClientShutdown()

    def get_state(self):
        """Get the state of the MessageSender and its underlying Link.

        :rtype: ~uamqp.constants.MessageSenderState
        """
        try:
            raise self._error
        except TypeError:
            pass
        except Exception as e:
            _logger.warning("%r", e)
            raise
        return self._state

    def work(self):
        """Update the link status."""
        self._link.do_work()

    def destroy(self):
        """Close both the Sender and the Link. Clean up any C objects."""
        self._sender.destroy()
        self._link.destroy()

    def open(self):
        """Open the MessageSender in order to start processing messages.

        :raises: ~uamqp.errors.AMQPConnectionError if the Sender raises
         an error on opening. This can happen if the target URI is invalid
         or the credentials are rejected.
        """
        try:
            self._sender.open()
        except ValueError:
            raise errors.AMQPConnectionError(
                "Failed to open Message Sender. "
                "Please confirm credentials and target URI.")

    def close(self):
        """Close the sender, leaving the link intact."""
        self._sender.close()

    def send(self, message, callback, timeout=0):
        """Add a single message to the internal pending queue to be processed
        by the Connection without waiting for it to be sent.

        :param message: The message to send.
        :type message: ~uamqp.message.Message
        :param callback: The callback to be run once a disposition is received
         in receipt of the message. The callback must take three arguments, the message,
         the send result and the optional delivery condition (exception).
        :type callback:
         callable[~uamqp.message.Message, ~uamqp.constants.MessageSendResult, ~uamqp.errors.MessageException]
        :param timeout: An expiry time for the message added to the queue. If the
         message is not sent within this timeout it will be discarded with an error
         state. If set to 0, the message will not expire. The default is 0.
        """
        # pylint: disable=protected-access
        try:
            raise self._error
        except TypeError:
            pass
        except Exception as e:
            _logger.warning("%r", e)
            raise
        c_message = message.get_message()
        message._on_message_sent = callback
        try:
            self._session._connection.lock(timeout=-1)
            return self._sender.send(c_message, timeout, message)
        finally:
            self._session._connection.release()

    def on_state_changed(self, previous_state, new_state):
        """Callback called whenever the underlying Sender undergoes a change
        of state. This function can be overridden.

        :param previous_state: The previous Sender state.
        :type previous_state: ~uamqp.constants.MessageSenderState
        :param new_state: The new Sender state.
        :type new_state: ~uamqp.constants.MessageSenderState
        """
        # pylint: disable=protected-access
        _logger.info("Message sender %r state changed from %r to %r on connection: %r",
                     self.name, previous_state, new_state, self._session._connection.container_id)
        self._state = new_state

    @property
    def send_settle_mode(self):
        return self._link.send_settle_mode

    @send_settle_mode.setter
    def send_settle_mode(self, value):
        self._link.send_settle_mode = value.value

    @property
    def receive_settle_mode(self):
        return self._link.receive_settle_mode

    @receive_settle_mode.setter
    def receive_settle_mode(self, value):
        self._link.receive_settle_mode = value.value

    @property
    def max_message_size(self):
        return self._link.max_message_size

    @max_message_size.setter
    def max_message_size(self, value):
        self._link.max_message_size = int(value)



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

# pylint: disable=super-init-not-called,no-self-use

import datetime
import logging
import time

from uamqp import Session, c_uamqp, compat, constants, errors, utils
from uamqp.constants import TransportType

from .common import _SASL, AMQPAuth

_logger = logging.getLogger(__name__)


class TokenRetryPolicy(object):
    """Retry policy for sending authentication tokens
    for CBS authentication.

    :param retries: The number of retry attempts for a failed
     PUT token request. The default is 3. This is exclusive of
     the initial attempt.
    :type retries: int
    :param backoff: The time in miliseconds to wait between
     retry attempts.
    :type backoff: int
    """

    def __init__(self, retries=3, backoff=0):
        self.retries = retries
        self.backoff = float(backoff)/1000


class CBSAuthMixin(object):
    """Mixin to handle sending and refreshing CBS auth tokens."""

    def update_token(self):
        """Update a token that is about to expire. This is specific
        to a particular token type, and therefore must be implemented
        in a child class.
        """
        raise errors.TokenExpired(
            "Unable to refresh token - no refresh logic implemented.")

    def create_authenticator(self, connection, debug=False, **kwargs):
        """Create the AMQP session and the CBS channel with which
        to negotiate the token.

        :param connection: The underlying AMQP connection on which
         to create the session.
        :type connection: ~uamqp.connection.Connection
        :param debug: Whether to emit network trace logging events for the
         CBS session. Default is `False`. Logging events are set at INFO level.
        :type debug: bool
        :rtype: uamqp.c_uamqp.CBSTokenAuth
        """
        self._connection = connection
        self._session = Session(connection, **kwargs)

        try:
            self._cbs_auth = c_uamqp.CBSTokenAuth(
                self.audience,
                self.token_type,
                self.token,
                int(self.expires_at),
                self._session._session,  # pylint: disable=protected-access
                self.timeout,
                self._connection.container_id,
                self._refresh_window
            )
            self._cbs_auth.set_trace(debug)
        except ValueError:
            self._session.destroy()
            raise errors.AMQPConnectionError(
                "Unable to open authentication session on connection {}.\n"
                "Please confirm target hostname exists: {}".format(connection.container_id, connection.hostname))
        return self._cbs_auth

    def close_authenticator(self):
        """Close the CBS auth channel and session."""
        _logger.info("Shutting down CBS session on connection: %r.", self._connection.container_id)
        try:
            _logger.debug("Unlocked CBS to close on connection: %r.", self._connection.container_id)
            self._cbs_auth.destroy()
            _logger.info("Auth closed, destroying session on connection: %r.", self._connection.container_id)
            self._session.destroy()
        finally:
            _logger.info("Finished shutting down CBS session on connection: %r.", self._connection.container_id)

    def handle_token(self):
        """This function is called periodically to check the status of the current
        token if there is one, and request a new one if needed.
        If the token request fails, it will be retried according to the retry policy.
        A token refresh will be attempted if the token will expire soon.

        This function will return a tuple of two booleans. The first represents whether
        the token authentication has not completed within it's given timeout window. The
        second indicates whether the token negotiation is still in progress.

        :raises: ~uamqp.errors.AuthenticationException if the token authentication fails.
        :raises: ~uamqp.errors.TokenExpired if the token has expired and cannot be
         refreshed.
        :rtype: tuple[bool, bool]
        """
        # pylint: disable=protected-access
        timeout = False
        in_progress = False
        try:
            self._connection.lock()
            if self._connection._closing or self._connection._error:
                return timeout, in_progress
            auth_status = self._cbs_auth.get_status()
            auth_status = constants.CBSAuthStatus(auth_status)
            if auth_status == constants.CBSAuthStatus.Error:
                if self.retries >= self._retry_policy.retries:  # pylint: disable=no-member
                    _logger.warning("Authentication Put-Token failed. Retries exhausted.")
                    raise errors.TokenAuthFailure(*self._cbs_auth.get_failure_info())
                error_code, error_description = self._cbs_auth.get_failure_info()
                _logger.info("Authentication status: %r, description: %r", error_code, error_description)
                _logger.info("Authentication Put-Token failed. Retrying.")
                self.retries += 1  # pylint: disable=no-member
                time.sleep(self._retry_policy.backoff)
                self._cbs_auth.authenticate()
                in_progress = True
            elif auth_status == constants.CBSAuthStatus.Failure:
                raise errors.AuthenticationException("Failed to open CBS authentication link.")
            elif auth_status == constants.CBSAuthStatus.Expired:
                raise errors.TokenExpired("CBS Authentication Expired.")
            elif auth_status == constants.CBSAuthStatus.Timeout:
                timeout = True
            elif auth_status == constants.CBSAuthStatus.InProgress:
                in_progress = True
            elif auth_status == constants.CBSAuthStatus.RefreshRequired:
                _logger.info("Token on connection %r will expire soon - attempting to refresh.",
                             self._connection.container_id)
                self.update_token()
                if self.token != self._prev_token:
                    self._cbs_auth.refresh(self.token, int(self.expires_at))
                else:
                    _logger.info(
                        "The newly acquired token on connection %r is the same as the previous one,"
                        " will keep attempting to refresh",
                        self._connection.container_id
                    )
            elif auth_status == constants.CBSAuthStatus.Idle:
                self._cbs_auth.authenticate()
                in_progress = True
            elif auth_status != constants.CBSAuthStatus.Ok:
                raise errors.AuthenticationException("Invalid auth state.")
        except compat.TimeoutException:
            _logger.debug("CBS auth timed out while waiting for lock acquisition.")
            return None, None
        except ValueError as e:
            raise errors.AuthenticationException(
                "Token authentication failed: {}".format(e))
        finally:
            self._connection.release()
        return timeout, in_progress

    def _set_expiry(self, expires_at, expires_in):
        if not expires_at and not expires_in:
            raise ValueError("Must specify either 'expires_at' or 'expires_in'.")
        if not expires_at:
            expires_at = time.time() + expires_in.seconds
        else:
            expires_in_seconds = expires_at - time.time()
            if expires_in_seconds < 1:
                raise ValueError("Token has already expired.")
            expires_in = datetime.timedelta(seconds=expires_in_seconds)
        return expires_at, expires_in


class SASTokenAuth(AMQPAuth, CBSAuthMixin):
    """CBS authentication using SAS tokens.

    :param audience: The token audience field. For SAS tokens
     this is usually the URI.
    :type audience: str or bytes
    :param uri: The AMQP endpoint URI. This must be provided as
     a decoded string.
    :type uri: str
    :param token: The SAS token.
    :type token: str or bytes.
    :param expires_in: The total remaining seconds until the token
     expires.
    :type expires_in: ~datetime.timedelta
    :param expires_at: The timestamp at which the SAS token will expire
     formatted as seconds since epoch.
    :type expires_at: float
    :param username: The SAS token username, also referred to as the key
     name or policy name. This can optionally be encoded into the URI.
    :type username: str
    :param password: The SAS token password, also referred to as the key.
     This can optionally be encoded into the URI.
    :type password: str
    :param port: The TLS port - default for AMQP is 5671.
    :type port: int
    :param timeout: The timeout in seconds in which to negotiate the token.
     The default value is 10 seconds.
    :type timeout: float
    :param retry_policy: The retry policy for the PUT token request. The default
     retry policy has 3 retries.
    :type retry_policy: ~uamqp.authentication.cbs_auth.TokenRetryPolicy
    :param verify: The path to a user-defined certificate.
    :type verify: str
    :param token_type: The type field of the token request.
     Default value is `b"servicebus.windows.net:sastoken"`.
    :type token_type: bytes
    :param http_proxy: HTTP proxy configuration. This should be a dictionary with
     the following keys present: 'proxy_hostname' and 'proxy_port'. Additional optional
     keys are 'username' and 'password'.
    :type http_proxy: dict
    :param transport_type: The transport protocol type - default is ~uamqp.TransportType.Amqp.
     ~uamqp.TransportType.AmqpOverWebsocket is applied when http_proxy is set or the
     transport type is explicitly requested.
    :type transport_type: ~uamqp.TransportType
    :param encoding: The encoding to use if hostname is provided as a str.
     Default is 'UTF-8'.
    :type encoding: str
    :keyword int refresh_window: The time in seconds before the token expiration
     time to start the process of token refresh.
     Default value is 10% of the remaining seconds until the token expires.
    """

    def __init__(self, audience, uri, token,
                 expires_in=None,
                 expires_at=None,
                 username=None,
                 password=None,
                 port=None,
                 timeout=10,
                 retry_policy=TokenRetryPolicy(),
                 verify=None,
                 token_type=b"servicebus.windows.net:sastoken",
                 http_proxy=None,
                 transport_type=TransportType.Amqp,
                 encoding='UTF-8',
                 **kwargs):  # pylint: disable=no-member
        self._retry_policy = retry_policy
        self._encoding = encoding
        self._refresh_window = kwargs.pop("refresh_window", 0)
        self._prev_token = None
        self.uri = uri
        parsed = compat.urlparse(uri)  # pylint: disable=no-member

        self.cert_file = verify
        self.hostname = (kwargs.get("custom_endpoint_hostname") or parsed.hostname).encode(self._encoding)
        self.username = compat.unquote_plus(parsed.username) if parsed.username else None  # pylint: disable=no-member
        self.password = compat.unquote_plus(parsed.password) if parsed.password else None  # pylint: disable=no-member

        self.username = username or self.username
        self.password = password or self.password
        self.audience = self._encode(audience)
        self.token_type = self._encode(token_type)
        self.token = self._encode(token)
        self.expires_at, self.expires_in = self._set_expiry(expires_at, expires_in)
        self.timeout = timeout
        self.retries = 0
        self.sasl = _SASL()
        self.set_io(self.hostname, port, http_proxy, transport_type)

    def update_token(self):
        """If a username and password are present - attempt to use them to
        request a fresh SAS token.
        """
        if not self.username or not self.password:
            raise errors.TokenExpired("Unable to refresh token - no username or password.")
        encoded_uri = compat.quote_plus(self.uri).encode(self._encoding)  # pylint: disable=no-member
        encoded_key = compat.quote_plus(self.username).encode(self._encoding)  # pylint: disable=no-member
        self.expires_at = time.time() + self.expires_in.seconds
        self._prev_token = self.token
        self.token = utils.create_sas_token(
            encoded_key,
            self.password.encode(self._encoding),
            encoded_uri,
            self.expires_in)

    @classmethod
    def from_shared_access_key(
            cls,
            uri,
            key_name,
            shared_access_key,
            expiry=None,
            port=None,
            timeout=10,
            retry_policy=TokenRetryPolicy(),
            verify=None,
            http_proxy=None,
            transport_type=TransportType.Amqp,
            encoding='UTF-8',
            **kwargs):
        """Attempt to create a CBS token session using a Shared Access Key such
        as is used to connect to Azure services.

        :param uri: The AMQP endpoint URI. This must be provided as
         a decoded string.
        :type uri: str
        :param key_name: The SAS token username, also referred to as the key
         name or policy name.
        :type key_name: str
        :param shared_access_key: The SAS token password, also referred to as the key.
        :type shared_access_key: str
        :param expiry: The lifetime in seconds for the generated token. Default is 1 hour.
        :type expiry: int
        :param port: The TLS port - default for AMQP is 5671.
        :type port: int
        :param timeout: The timeout in seconds in which to negotiate the token.
         The default value is 10 seconds.
        :type timeout: float
        :param retry_policy: The retry policy for the PUT token request. The default
         retry policy has 3 retries.
        :type retry_policy: ~uamqp.authentication.cbs_auth.TokenRetryPolicy
        :param verify: The path to a user-defined certificate.
        :type verify: str
        :param http_proxy: HTTP proxy configuration. This should be a dictionary with
         the following keys present: 'proxy_hostname' and 'proxy_port'. Additional optional
         keys are 'username' and 'password'.
        :type http_proxy: dict
        :param transport_type: The transport protocol type - default is ~uamqp.TransportType.Amqp.
         ~uamqp.TransportType.AmqpOverWebsocket is applied when http_proxy is set or the
         transport type is explicitly requested.
        :type transport_type: ~uamqp.TransportType
        :param encoding: The encoding to use if hostname is provided as a str.
         Default is 'UTF-8'.
        :type encoding: str
        :keyword int refresh_window: The time in seconds before the token expiration
        time to start the process of token refresh.
        Default value is 10% of the remaining seconds until the token expires.
        """
        expires_in = datetime.timedelta(seconds=expiry or constants.AUTH_EXPIRATION_SECS)
        encoded_uri = compat.quote_plus(uri).encode(encoding)  # pylint: disable=no-member
        encoded_key = compat.quote_plus(key_name).encode(encoding)  # pylint: disable=no-member
        expires_at = time.time() + expires_in.seconds
        token = utils.create_sas_token(
            encoded_key,
            shared_access_key.encode(encoding),
            encoded_uri,
            expires_in)
        return cls(
            uri, uri, token,
            expires_in=expires_in,
            expires_at=expires_at,
            username=key_name,
            password=shared_access_key,
            port=port,
            timeout=timeout,
            retry_policy=retry_policy,
            verify=verify,
            http_proxy=http_proxy,
            transport_type=transport_type,
            encoding=encoding,
            custom_endpoint_hostname=kwargs.pop("custom_endpoint_hostname", None))


class JWTTokenAuth(AMQPAuth, CBSAuthMixin):
    """CBS authentication using JWT tokens.

    :param audience: The token audience field. For JWT tokens
     this is usually the URI.
    :type audience: str or bytes
    :param uri: The AMQP endpoint URI. This must be provided as
     a decoded string.
    :type uri: str
    :param get_token: The callback function used for getting and refreshing
     tokens. It should return a valid jwt token each time it is called.
    :type get_token: callable object
    :param expires_in: The total remaining seconds until the token
     expires - default for JWT token generated by AAD is 3600s (1 hour).
    :type expires_in: ~datetime.timedelta
    :param expires_at: The timestamp at which the JWT token will expire
     formatted as seconds since epoch.
    :type expires_at: float
    :param port: The TLS port - default for AMQP is 5671.
    :type port: int
    :param timeout: The timeout in seconds in which to negotiate the token.
     The default value is 10 seconds.
    :type timeout: float
    :param retry_policy: The retry policy for the PUT token request. The default
     retry policy has 3 retries.
    :type retry_policy: ~uamqp.authentication.cbs_auth.TokenRetryPolicy
    :param verify: The path to a user-defined certificate.
    :type verify: str
    :param token_type: The type field of the token request.
     Default value is `b"jwt"`.
    :type token_type: bytes
    :param http_proxy: HTTP proxy configuration. This should be a dictionary with
     the following keys present: 'proxy_hostname' and 'proxy_port'. Additional optional
     keys are 'username' and 'password'.
    :type http_proxy: dict
    :param transport_type: The transport protocol type - default is ~uamqp.TransportType.Amqp.
     ~uamqp.TransportType.AmqpOverWebsocket is applied when http_proxy is set or the
     transport type is explicitly requested.
    :type transport_type: ~uamqp.TransportType
    :param encoding: The encoding to use if hostname is provided as a str.
     Default is 'UTF-8'.
    :type encoding: str
    :keyword int refresh_window: The time in seconds before the token expiration
     time to start the process of token refresh.
     Default value is 10% of the remaining seconds until the token expires.
    """

    def __init__(self, audience, uri,
                 get_token,
                 expires_in=datetime.timedelta(seconds=constants.AUTH_EXPIRATION_SECS),
                 expires_at=None,
                 port=None,
                 timeout=10,
                 retry_policy=TokenRetryPolicy(),
                 verify=None,
                 token_type=b"jwt",
                 http_proxy=None,
                 transport_type=TransportType.Amqp,
                 encoding='UTF-8',
                 **kwargs):  # pylint: disable=no-member
        self._retry_policy = retry_policy
        self._encoding = encoding
        self._refresh_window = kwargs.pop("refresh_window", 0)
        self._prev_token = None
        self.uri = uri
        parsed = compat.urlparse(uri)  # pylint: disable=no-member

        self.cert_file = verify
        self.hostname = (kwargs.get("custom_endpoint_hostname") or parsed.hostname).encode(self._encoding)

        if not get_token or not callable(get_token):
            raise ValueError("get_token must be a callable object.")

        self.get_token = get_token
        self.audience = self._encode(audience)
        self.token_type = self._encode(token_type)
        self.token = None
        self.expires_at, self.expires_in = self._set_expiry(expires_at, expires_in)
        self.timeout = timeout
        self.retries = 0
        self.sasl = _SASL()
        self.set_io(self.hostname, port, http_proxy, transport_type)

    def create_authenticator(self, connection, debug=False, **kwargs):
        self.update_token()
        return super(JWTTokenAuth, self).create_authenticator(connection, debug, **kwargs)

    def update_token(self):
        access_token = self.get_token()
        self.expires_at = access_token.expires_on
        self._prev_token = self.token
        self.token = self._encode(access_token.token)



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

from enum import Enum

from uamqp import c_uamqp

DEFAULT_AMQPS_PORT = 5671
DEFAULT_AMQP_WSS_PORT = 443
AUTH_EXPIRATION_SECS = c_uamqp.AUTH_EXPIRATION_SECS
AUTH_REFRESH_SECS = c_uamqp.AUTH_REFRESH_SECS
BATCH_MESSAGE_FORMAT = c_uamqp.AMQP_BATCH_MESSAGE_FORMAT
MAX_FRAME_SIZE_BYTES = c_uamqp.MAX_FRAME_SIZE_BYTES
MAX_MESSAGE_LENGTH_BYTES = c_uamqp.MAX_MESSAGE_LENGTH_BYTES
STRING_FILTER = b"apache.org:selector-filter:string"
OPERATION = b"operation"
READ_OPERATION = b"READ"
MGMT_TARGET = b"$management"

# Deprecated - will be removed in future versions
MESSAGE_SEND_RETRIES = 3
ERROR_CONNECTION_REDIRECT = b"amqp:connection:redirect"
ERROR_LINK_REDIRECT = b"amqp:link:redirect"


class MessageState(Enum):
    WaitingToBeSent = 0
    WaitingForSendAck = 1
    SendComplete = 2
    SendFailed = 3
    ReceivedUnsettled = 4
    ReceivedSettled = 5


DONE_STATES = (MessageState.SendComplete, MessageState.SendFailed)
RECEIVE_STATES = (MessageState.ReceivedSettled, MessageState.ReceivedUnsettled)
PENDING_STATES = (MessageState.WaitingForSendAck, MessageState.WaitingToBeSent)


# Error Codes
class ErrorCodes(Enum):
    InternalServerError = b"amqp:internal-error"
    IllegalState = b"amqp:illegal-state"
    DecodeError = b"amqp:decode-error"
    NotFound = b"amqp:not-found"
    NotImplemented = b"amqp:not-implemented"
    NotAllowed = b"amqp:not-allowed"
    InvalidField = b"amqp:invalid-field"
    ResourceLocked = b"amqp:resource-locked"
    ResourceDeleted = b"amqp:resource-deleted"
    UnauthorizedAccess = b"amqp:unauthorized-access"
    FrameSizeTooSmall = b"amqp:frame-size-too-small"
    ResourceLimitExceeded = b"amqp:resource-limit-exceeded"
    PreconditionFailed = b"amqp:precondition-failed"
    ConnectionRedirect = b"amqp:connection:redirect"
    ConnectionCloseForced = b"amqp:connection:forced"
    ConnectionFramingError = b"amqp:connection:framing-error"
    SessionWindowViolation = b"amqp:session:window-violation"
    SessionErrantLink = b"amqp:session:errant-link"
    SessionHandleInUse = b"amqp:session:handle-in-use"
    SessionUnattachedHandle = b"amqp:session:unattached-handle"
    LinkRedirect = b"amqp:link:redirect"
    LinkStolen = b"amqp:link:stolen"
    LinkDetachForced = b"amqp:link:detach-forced"
    LinkTransferLimitExceeded = b"amqp:link:transfer-limit-exceeded"
    LinkMessageSizeExceeded = b"amqp:link:message-size-exceeded"
    ClientError = b"amqp:client-error"
    UnknownError = b"amqp:unknown-error"
    VendorError = b"amqp:vendor-error"


class MessageReceiverState(Enum):
    Idle = c_uamqp.MESSAGE_RECEIVER_STATE_IDLE
    Opening = c_uamqp.MESSAGE_RECEIVER_STATE_OPENING
    Open = c_uamqp.MESSAGE_RECEIVER_STATE_OPEN
    Closing = c_uamqp.MESSAGE_RECEIVER_STATE_CLOSING
    Error = c_uamqp.MESSAGE_RECEIVER_STATE_ERROR


class MessageSendResult(Enum):
    Ok = c_uamqp.MESSAGE_SEND_OK
    Error = c_uamqp.MESSAGE_SEND_ERROR
    Timeout = c_uamqp.MESSAGE_SEND_TIMEOUT
    Cancelled = c_uamqp.MESSAGE_SEND_CANCELLED


class MessageSenderState(Enum):
    Idle = c_uamqp.MESSAGE_SENDER_STATE_IDLE
    Opening = c_uamqp.MESSAGE_SENDER_STATE_OPENING
    Open = c_uamqp.MESSAGE_SENDER_STATE_OPEN
    Closing = c_uamqp.MESSAGE_SENDER_STATE_CLOSING
    Error = c_uamqp.MESSAGE_SENDER_STATE_ERROR


class ManagementLinkState(Enum):
    Ok = c_uamqp.AMQP_MANAGEMENT_OPEN_OK
    Error = c_uamqp.AMQP_MANAGEMENT_OPEN_ERROR
    Cancelled = c_uamqp.AMQP_MANAGEMENT_OPEN_CANCELLED


class ManagementOperationResult(Enum):
    Ok = c_uamqp.AMQP_MANAGEMENT_EXECUTE_OPERATION_OK
    Error = c_uamqp.AMQP_MANAGEMENT_EXECUTE_OPERATION_ERROR
    BadStatus = c_uamqp.AMQP_MANAGEMENT_EXECUTE_OPERATION_FAILED_BAD_STATUS
    Closed = c_uamqp.AMQP_MANAGEMENT_EXECUTE_OPERATION_INSTANCE_CLOSED


class Role(Enum):
    Sender = c_uamqp.ROLE_SENDER
    Receiver = c_uamqp.ROLE_RECEIVER


class SenderSettleMode(Enum):
    Unsettled = c_uamqp.SENDER_SETTLE_MODE_UNSETTLED
    Settled = c_uamqp.SENDER_SETTLE_MODE_SETTLED
    Mixed = c_uamqp.SENDER_SETTLE_MODE_MIXED


class ReceiverSettleMode(Enum):
    PeekLock = c_uamqp.RECEIVER_SETTLE_MODE_PEEKLOCK
    ReceiveAndDelete = c_uamqp.RECEIVER_SETTLE_MODE_RECEIVEANDDELETE


class CBSOperationResult(Enum):
    Ok = c_uamqp.CBS_OPERATION_RESULT_OK
    Error = c_uamqp.CBS_OPERATION_RESULT_CBS_ERROR
    Failed = c_uamqp.CBS_OPERATION_RESULT_OPERATION_FAILED
    Closed = c_uamqp.CBS_OPERATION_RESULT_INSTANCE_CLOSED


class CBSOpenState(Enum):
    Ok = c_uamqp.CBS_OPEN_COMPLETE_OK
    Error = c_uamqp.CBS_OPEN_COMPLETE_ERROR
    Cancelled = c_uamqp.CBS_OPEN_COMPLETE_CANCELLED


class CBSAuthStatus(Enum):
    Ok = c_uamqp.AUTH_STATUS_OK
    Idle = c_uamqp.AUTH_STATUS_IDLE
    InProgress = c_uamqp.AUTH_STATUS_IN_PROGRESS
    Timeout = c_uamqp.AUTH_STATUS_TIMEOUT
    RefreshRequired = c_uamqp.AUTH_STATUS_REFRESH_REQUIRED
    Expired = c_uamqp.AUTH_STATUS_EXPIRED
    Error = c_uamqp.AUTH_STATUS_ERROR
    Failure = c_uamqp.AUTH_STATUS_FAILURE


class MgmtExecuteResult(Enum):
    Ok = c_uamqp.AMQP_MANAGEMENT_EXECUTE_OPERATION_OK
    Error = c_uamqp.AMQP_MANAGEMENT_EXECUTE_OPERATION_ERROR
    Failed = c_uamqp.AMQP_MANAGEMENT_EXECUTE_OPERATION_FAILED_BAD_STATUS
    Closed = c_uamqp.AMQP_MANAGEMENT_EXECUTE_OPERATION_INSTANCE_CLOSED


class MgmtOpenStatus(Enum):
    Ok = c_uamqp.AMQP_MANAGEMENT_OPEN_OK
    Error = c_uamqp.AMQP_MANAGEMENT_OPEN_ERROR
    Cancelled = c_uamqp.AMQP_MANAGEMENT_OPEN_CANCELLED


class TransportType(Enum):
    """Transport type
    The underlying transport protocol type:
     Amqp: AMQP over the default TCP transport protocol, it uses port 5671.
     AmqpOverWebsocket: Amqp over the Web Sockets transport protocol, it uses
     port 443.
    """
    Amqp = 1
    AmqpOverWebsocket = 2


class MessageBodyType(Enum):
    """AMQP message body type
    The body of an amqp message consists of either: one or more data sections, one or more amqp-sequence sections,
    or a single amqp-value section:
    `DataType`: The body consists of one or more data sections and each section contains opaque binary data.
    `SequenceType`: The body consists of one or more sequence sections and each section contains an arbitrary
     number of structured data elements.
    `ValueType`: The body consists of one amqp-value section and the section contains a single AMQP value.

    Please refer to the AMQP spec:
    http://docs.oasis-open.org/amqp/core/v1.0/os/amqp-core-messaging-v1.0-os.html#section-message-format
    for further information on the message body type.
    """
    Data = c_uamqp.MessageBodyType.DataType
    Value = c_uamqp.MessageBodyType.ValueType
    Sequence = c_uamqp.MessageBodyType.SequenceType


BODY_TYPE_C_PYTHON_MAP = {
    c_uamqp.MessageBodyType.DataType.value: MessageBodyType.Data,
    c_uamqp.MessageBodyType.SequenceType.value: MessageBodyType.Sequence,
    c_uamqp.MessageBodyType.ValueType.value: MessageBodyType.Value
}



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

from libc cimport stdint

cimport c_session
cimport c_strings
cimport c_async_operation


cdef enum AUTH_STATUS:
    AUTH_STATUS_OK,
    AUTH_STATUS_IDLE,
    AUTH_STATUS_IN_PROGRESS,
    AUTH_STATUS_TIMEOUT,
    AUTH_STATUS_REFRESH_REQUIRED,
    AUTH_STATUS_EXPIRED,
    AUTH_STATUS_ERROR,
    AUTH_STATUS_FAILURE


cdef extern from "azure_uamqp_c/cbs.h":

    cdef enum CBS_OPERATION_RESULT_TAG:
        CBS_OPERATION_RESULT_OK,
        CBS_OPERATION_RESULT_CBS_ERROR,
        CBS_OPERATION_RESULT_OPERATION_FAILED,
        CBS_OPERATION_RESULT_INSTANCE_CLOSED

    cdef enum CBS_OPEN_COMPLETE_RESULT_TAG:
        CBS_OPEN_OK,
        CBS_OPEN_ERROR,
        CBS_OPEN_CANCELLED

    ctypedef struct CBS_HANDLE:
        pass

    ctypedef void (*ON_CBS_OPEN_COMPLETE)(void* context, CBS_OPEN_COMPLETE_RESULT_TAG open_complete_result)
    ctypedef void (*ON_CBS_ERROR)(void* context)
    ctypedef void (*ON_CBS_OPERATION_COMPLETE)(void* context, CBS_OPERATION_RESULT_TAG complete_result, unsigned int status_code, const char* status_description)

    CBS_HANDLE cbs_create(c_session.SESSION_HANDLE session)
    void cbs_destroy(CBS_HANDLE cbs)
    int cbs_open_async(CBS_HANDLE cbs, ON_CBS_OPEN_COMPLETE on_cbs_open_complete, void* on_cbs_open_complete_context, ON_CBS_ERROR on_cbs_error, void* on_cbs_error_context)
    int cbs_close(CBS_HANDLE cbs)
    c_async_operation.ASYNC_OPERATION_HANDLE cbs_put_token_async(CBS_HANDLE cbs, const char* type, const char* audience, const char* token, ON_CBS_OPERATION_COMPLETE on_cbs_put_token_complete, void* on_cbs_put_token_complete_context)
    int cbs_delete_token_async(CBS_HANDLE cbs, const char* type, const char* audience, ON_CBS_OPERATION_COMPLETE on_cbs_delete_token_complete, void* on_cbs_delete_token_complete_context)
    int cbs_set_trace(CBS_HANDLE cbs, bint trace_on)



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

# C imports
cimport c_tlsio
cimport c_xio


DEFAULT_PORT = 5671


cdef class TLSIOConfig(object):

    cdef c_tlsio.TLSIO_CONFIG _c_value

    def __cinit__(self):
        self._c_value = c_tlsio.TLSIO_CONFIG(NULL, DEFAULT_PORT, NULL, NULL)

    @property
    def hostname(self):
        return self._c_value.hostname

    @hostname.setter
    def hostname(self, const char* value):
        self._c_value.hostname = value

    @property
    def port(self):
        return self._c_value.port

    @port.setter
    def port(self, int port):
        self._c_value.port = port

    cpdef set_proxy_config(self, HTTPProxyConfig underlying_io_parameters):
        cdef const c_xio.IO_INTERFACE_DESCRIPTION* proxy_io
        proxy_io = c_tlsio.http_proxy_io_get_interface_description()
        self._c_value.underlying_io_interface = proxy_io
        self._c_value.underlying_io_parameters = &underlying_io_parameters._c_value


cdef class HTTPProxyConfig(object):

    cdef c_tlsio.HTTP_PROXY_IO_CONFIG _c_value

    def __cinit__(self):
        self._c_value = c_tlsio.HTTP_PROXY_IO_CONFIG(NULL, 0, NULL, 0, NULL, NULL)

    @property
    def hostname(self):
        return self._c_value.hostname

    @hostname.setter
    def hostname(self, const char* value):
        self._c_value.hostname = value

    @property
    def port(self):
        return self._c_value.port

    @port.setter
    def port(self, int port):
        self._c_value.port = port

    @property
    def proxy_hostname(self):
        return self._c_value.proxy_hostname

    @proxy_hostname.setter
    def proxy_hostname(self, const char* value):
        self._c_value.proxy_hostname = value

    @property
    def proxy_port(self):
        return self._c_value.proxy_port

    @proxy_port.setter
    def proxy_port(self, int port):
        self._c_value.proxy_port = port

    @property
    def username(self):
        return self._c_value.username

    @username.setter
    def username(self, const char* value):
        self._c_value.username = value

    @property
    def password(self):
        return self._c_value.password

    @password.setter
    def password(self, const char* value):
        self._c_value.password = value



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

# pylint: disable=no-name-in-module,unused-import,redefined-builtin,import-error,undefined-variable,ungrouped-imports

import sys

from uamqp import errors

PY2 = (2, 7) <= sys.version_info < (3, 0)
PY3 = (3, 4) <= sys.version_info < (4, 0)


if PY3:
    import queue
    from urllib.parse import urlparse, unquote_plus, quote_plus

    TimeoutException = TimeoutError

    builtin_str = str
    str = str
    bytes = bytes
    basestring = (str, bytes,)
    numeric_types = (int, float,)
    integer_types = (int,)

    long = int

elif PY2:
    import Queue as queue
    from urlparse import urlparse
    from urllib import unquote_plus, quote_plus

    TimeoutException = errors.ClientTimeout

    builtin_str = str
    bytes = str
    str = unicode
    basestring = basestring
    numeric_types = (int, long, float,)
    integer_types = (int, long,)

    long = long



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

# Python imports
import logging

# C imports
cimport c_xio
cimport c_sasl_mechanism


_logger = logging.getLogger(__name__)


cpdef _get_sasl_mechanism_interface():
    cdef const c_sasl_mechanism.SASL_MECHANISM_INTERFACE_DESCRIPTION* interface
    interface = c_sasl_mechanism.saslmssbcbs_get_interface()
    if <void*>interface == NULL:
        raise ValueError("Failed to create SASL CBS Interface description")
    desc = SASLMechanismInterfaceDescription()
    desc.wrap(interface)
    return desc


cpdef saslanonymous_get_interface():
    cdef c_sasl_mechanism.SASL_MECHANISM_INTERFACE_DESCRIPTION* io_desc
    io_desc = c_sasl_mechanism.saslanonymous_get_interface()
    if <void*>io_desc == NULL:
        raise ValueError("Failed to create SASL Anonymous Interface description")

    interface = SASLMechanismInterfaceDescription()
    interface.wrap(io_desc)
    return interface


cpdef saslplain_get_interface():
    cdef c_sasl_mechanism.SASL_MECHANISM_INTERFACE_DESCRIPTION* io_desc
    io_desc = c_sasl_mechanism.saslplain_get_interface()
    if <void*>io_desc == NULL:
        raise ValueError("Failed to create SASL plain Interface description")

    interface = SASLMechanismInterfaceDescription()
    interface.wrap(io_desc)
    return interface


cpdef get_sasl_mechanism(SASLMechanismInterfaceDescription interface=None):
    if interface is None:
        interface = _get_sasl_mechanism_interface()
    sasl_mechanism = SASLMechanism()
    sasl_mechanism.create(interface)
    return sasl_mechanism


cpdef get_plain_sasl_mechanism(SASLMechanismInterfaceDescription interface, SASLPlainConfig parameters):
    _logger.debug("Creating SASL Mechanism")
    sasl_mechanism = SASLMechanism()
    sasl_mechanism.create_with_parameters(interface, &parameters._c_value)
    return sasl_mechanism



cdef class SASLMechanism(StructBase):

    cdef c_sasl_mechanism.SASL_MECHANISM_HANDLE _c_value

    def __cinit__(self):
        pass

    def __dealloc__(self):
        _logger.debug("Deallocating SASLMechanism")
        self.destroy()

    cdef _create(self):
        if <void*>self._c_value is NULL:
            self._memory_error()

    cpdef destroy(self):
        if <void*>self._c_value is not NULL:
            _logger.debug("Destroying SASLMechanism")
            c_sasl_mechanism.saslmechanism_destroy(self._c_value)
            self._c_value = <c_sasl_mechanism.SASL_MECHANISM_HANDLE>NULL

    cdef wrap(self, c_sasl_mechanism.SASL_MECHANISM_HANDLE value):
        self.destroy()
        self._c_value = value
        self._create()

    cdef create(self, SASLMechanismInterfaceDescription sasl_mechanism_interface_description):
        self.destroy()
        self._c_value = c_sasl_mechanism.saslmechanism_create(sasl_mechanism_interface_description._c_value, NULL)
        self._create()

    cdef create_with_parameters(self, SASLMechanismInterfaceDescription sasl_mechanism_interface_description, void *parameters):
        self.destroy()
        self._c_value = c_sasl_mechanism.saslmechanism_create(sasl_mechanism_interface_description._c_value, parameters)
        self._create()


cdef class SASLMechanismInterfaceDescription(object):

    cdef c_sasl_mechanism.SASL_MECHANISM_INTERFACE_DESCRIPTION* _c_value

    def __cinit__(self):
        pass

    cdef wrap(self, c_sasl_mechanism.SASL_MECHANISM_INTERFACE_DESCRIPTION* value):
        self._c_value = value


cdef class SASLClientIOConfig(object):

    cdef c_sasl_mechanism.SASLCLIENTIO_CONFIG _c_value
    cdef XIO _underlying_io

    def __cinit__(self, XIO underlying_io, SASLMechanism sasl_mechanism):
        if <void*>underlying_io._c_value is NULL:
            raise ValueError("UnderLying IO must not be NULL")
        if <void*>sasl_mechanism._c_value is NULL:
            raise ValueError("SASL Mechanism must not be NULL")

        self._underlying_io = underlying_io
        self._c_value = c_sasl_mechanism.SASLCLIENTIO_CONFIG(
            <c_xio.XIO_HANDLE>underlying_io._c_value,
            <c_sasl_mechanism.SASL_MECHANISM_HANDLE>sasl_mechanism._c_value
        )


cdef class SASLPlainConfig(object):

    cdef c_sasl_mechanism.SASL_PLAIN_CONFIG _c_value

    def __cinit__(self):
        self._c_value = c_sasl_mechanism.SASL_PLAIN_CONFIG(NULL, NULL, NULL)
        self._c_value.authcid = NULL
        self._c_value.passwd = NULL
        self._c_value.authzid = NULL

    @property
    def authcid(self):
        return self._c_value.authcid

    @authcid.setter
    def authcid(self, const char* value):
        self._c_value.authcid = value

    @property
    def passwd(self):
        return self._c_value.passwd

    @passwd.setter
    def passwd(self, const char* value):
        self._c_value.passwd = value

    @property
    def authzid(self):
        return self._c_value.authzid

    @authzid.setter
    def authzid(self, const char* value):
        self._c_value.authzid = value



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

# Python imports
import logging
import copy

# C imports
cimport c_amqp_definitions
cimport c_amqpvalue


_logger = logging.getLogger(__name__)


cpdef create_target():
    target = cTarget()
    return target


cdef target_factory(c_amqp_definitions.TARGET_HANDLE c_target):
    target = cTarget()
    target.wrap(c_target)
    return target


cdef class cTarget(StructBase):

    cdef c_amqp_definitions.TARGET_HANDLE _c_value

    def __cinit__(self):
        self._c_value = c_amqp_definitions.target_create()
        self._validate()

    def __dealloc__(self):
        _logger.debug("Deallocating cTarget")
        self.destroy()

    cdef _validate(self):
        if <void*>self._c_value is NULL:
            self._memory_error()

    cpdef destroy(self):
        if <void*>self._c_value is not NULL:
            _logger.debug("Destroying cTarget")
            c_amqp_definitions.target_destroy(self._c_value)
            self._c_value = <c_amqp_definitions.TARGET_HANDLE>NULL

    cdef wrap(self, c_amqp_definitions.TARGET_HANDLE value):
        self.destroy()
        self._c_value = value
        self._validate()

    @property
    def value(self):
        cdef c_amqpvalue.AMQP_VALUE _value
        _value = c_amqp_definitions.amqpvalue_create_target(self._c_value)
        if <void*>_value == NULL:
            self._null_error("Failed to create target.")
        return value_factory(_value)

    @property
    def address(self):
        cdef c_amqpvalue.AMQP_VALUE _value
        cdef c_amqpvalue.AMQP_VALUE cloned
        if c_amqp_definitions.target_get_address(self._c_value, &_value) != 0:
            self._value_error("Failed to get target address")
        if <void*>_value == NULL:
            return None
        cloned = c_amqpvalue.amqpvalue_clone(_value)
        if <void*>cloned == NULL:
            return None
        return value_factory(cloned).value

    @address.setter
    def address(self, AMQPValue value):
        if c_amqp_definitions.target_set_address(self._c_value, <c_amqpvalue.AMQP_VALUE>value._c_value) != 0:
            self._value_error("Failed to set target address")

    @property
    def durable(self):
        cdef c_amqp_definitions.terminus_durability _value
        if c_amqp_definitions.target_get_durable(self._c_value, &_value) != 0:
            self._value_error("Failed to get target durable")
        if <void*>_value == NULL:
            return None
        return _value

    @durable.setter
    def durable(self, c_amqp_definitions.terminus_durability value):
        if c_amqp_definitions.target_set_durable(self._c_value, value) != 0:
            self._value_error("Failed to set target durable")

    @property
    def expiry_policy(self):
        cdef c_amqp_definitions.terminus_expiry_policy _value
        if c_amqp_definitions.target_get_expiry_policy(self._c_value, &_value) != 0:
            self._value_error("Failed to get target expiry_policy")
        if <void*>_value == NULL:
            return None
        return _value

    @expiry_policy.setter
    def expiry_policy(self, c_amqp_definitions.terminus_expiry_policy value):
        if c_amqp_definitions.target_set_expiry_policy(self._c_value, value) != 0:
            self._value_error("Failed to set target expiry_policy")

    @property
    def timeout(self):
        cdef c_amqp_definitions.seconds _value
        if c_amqp_definitions.target_get_timeout(self._c_value, &_value) != 0:
            self._value_error("Failed to get target timeout")
        if <void*>_value == NULL:
            return None
        return _value

    @timeout.setter
    def timeout(self, c_amqp_definitions.seconds value):
        if c_amqp_definitions.target_set_timeout(self._c_value, value) != 0:
            self._value_error("Failed to set target timeout")

    @property
    def dynamic(self):
        cdef bint _value
        if c_amqp_definitions.target_get_dynamic(self._c_value, &_value) != 0:
            self._value_error("Failed to get target dynamic")
        if <void*>_value == NULL:
            return None
        return _value

    @dynamic.setter
    def dynamic(self, bint value):
        if c_amqp_definitions.target_set_dynamic(self._c_value, value) != 0:
            self._value_error("Failed to set target dynamic")

    @property
    def dynamic_node_properties(self):
        cdef c_amqp_definitions.node_properties _value
        if c_amqp_definitions.target_get_dynamic_node_properties(self._c_value, &_value) != 0:
            self._value_error("Failed to get target dynamic_node_properties")
        if <void*>_value == NULL:
            return None
        return annotations_factory(_value)

    @dynamic_node_properties.setter
    def dynamic_node_properties(self, cFields value):
        if c_amqp_definitions.target_set_dynamic_node_properties(self._c_value, <c_amqp_definitions.node_properties>value._c_value) != 0:
            self._value_error("Failed to set target dynamic_node_properties")
